"""
IEC 61672-1:2013
================

IEC 61672-1:2013 gives electroacoustical performance specifications for three 
kinds of sound measuring instruments [IEC61672]_:

- time-weighting sound level meters that measure exponential-time-weighted, frequency-weighted sound levels;
- integrating-averaging sound level meters that measure time-averaged, frequency-weighted sound levels; and
- integrating sound level meters that measure frequency-weighted sound exposure levels. 

.. [IEC61672] http://webstore.iec.ch/webstore/webstore.nsf/artnum/048669!opendocument

Weighting functions
*******************

.. autofunction:: weighting_function_a
.. autofunction:: weighting_function_c
.. autofunction:: weighting_function_z


Weighting systems
*****************

.. autofunction:: weighting_system_a
.. autofunction:: weighting_system_c
.. autofunction:: weighting_system_z



"""
from __future__ import division

import numpy as np
from scipy.signal import zpk2tf
from scipy.signal import lfilter, bilinear
from math import floor
from .iso_tr_25417_2007 import REFERENCE_PRESSURE

import io
import os
import pkgutil
import pandas as pd

WEIGHTING_DATA = pd.read_table(io.BytesIO(pkgutil.get_data('acoustics', os.path.join('data', 'iec_61672_1_2013.csv'))), sep=',', index_col=0)
"""DataFrame with indices, nominal frequencies and weighting values.
"""

NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES = np.array(WEIGHTING_DATA.nominal)
"""Nominal 1/3-octave frequencies. See table 3.
"""

NOMINAL_OCTAVE_CENTER_FREQUENCIES = np.array(WEIGHTING_DATA.nominal)[2::3]
"""Nominal 1/1-octave frequencies. Based on table 3.
"""

REFERENCE_FREQUENCY = 1000.0
"""Reference frequency. See table 3.
"""

EXACT_THIRD_OCTAVE_CENTER_FREQUENCIES = REFERENCE_FREQUENCY * 10.0**(0.01*(np.arange(10, 44)-30))
"""Exact third-octave center frequencies. See table 3.
"""

WEIGHTING_A = np.array(WEIGHTING_DATA.A)
"""Frequency weighting A. See table 3.
"""

WEIGHTING_C = np.array(WEIGHTING_DATA.C)
"""Frequency weighting C. See table 3.
"""

WEIGHTING_Z = np.array(WEIGHTING_DATA.Z)
"""Frequency weighting Z. See table 3.
"""

WEIGHTING_VALUES = {'A': WEIGHTING_A,
                    'C': WEIGHTING_C,
                    'Z': WEIGHTING_Z
                    }
"""Dictionary with weighting values 'A', 'C' and 'Z' weighting.
"""


FAST = 0.125
"""FAST time-constant.
"""

SLOW = 1.000
"""SLOW time-constant.
"""


def time_averaged_sound_level(pressure, sample_frequency, averaging_time, reference_pressure=REFERENCE_PRESSURE):
    """Time-averaged sound pressure level.
    
    :param pressure: Dynamic pressure.
    :param sample_frequency: Sample frequency.
    :param averaging_time: Averaging time.
    :param reference_pressure: Reference pressure.
    
    """
    levels = 10.0 * np.log10( average(pressure**2.0, sample_frequency, averaging_time) / reference_pressure**2.0)
    times = np.arange(levels.shape[-1]) * averaging_time
    return times, levels


def average(data, sample_frequency, averaging_time):
    """Average the sound pressure squared.
    
    :param data: Energetic quantity, e.g. :math:`p^2`.
    :param sample_frequency: Sample frequency.
    :param averaging_time: Averaging time.
    :returns: 
    
    Time weighting is applied by applying a low-pass filter with one real pole at :math:`-1/\\tau`.
    
    .. note:: Because :math:`f_s \\cdot t_i` is generally not an integer, samples are discarded. This results in a drift of samples for longer signals (e.g. 60 minutes at 44.1 kHz).
    
    """
    averaging_time = np.asarray(averaging_time)
    sample_frequency = np.asarray(sample_frequency)
    samples = data.shape[-1]
    n = np.floor(averaging_time * sample_frequency).astype(int)
    data = data[..., 0:n*(samples//n)] # Drop the tail of the signal.
    newshape = list(data.shape[0:-1])
    newshape.extend([-1, n])
    data = data.reshape(newshape)
    #data = data.reshape((-1, n))
    return data.mean(axis=-1)

def time_weighted_sound_level(pressure, sample_frequency, integration_time, reference_pressure=REFERENCE_PRESSURE):
    """Time-weighted sound pressure level.
    
    :param pressure: Dynamic pressure.
    :param sample_frequency: Sample frequency.
    :param integration_time: Integration time.
    :param reference_pressure: Reference pressure.
    """
    levels = 10.0 * np.log10( integrate(pressure**2.0, sample_frequency, integration_time) / reference_pressure**2.0)
    times = np.arange(levels.shape[-1]) * integration_time
    return times, levels

def integrate(data, sample_frequency, integration_time):
    """Integrate the sound pressure squared using exponential integration.
    
    :param data: Energetic quantity, e.g. :math:`p^2`.
    :param sample_frequency: Sample frequency.
    :param integration_time: Integration time.
    :returns: 
    
    Time weighting is applied by applying a low-pass filter with one real pole at :math:`-1/\\tau`.
    
    .. note:: Because :math:`f_s \\cdot t_i` is generally not an integer, samples are discarded. This results in a drift of samples for longer signals (e.g. 60 minutes at 44.1 kHz).
    
    """
    integration_time = np.asarray(integration_time)
    sample_frequency = np.asarray(sample_frequency)
    samples = data.shape[-1]
    b, a  = zpk2tf([1.0], [1.0, integration_time], [1.0])
    b, a = bilinear(b, a, fs=sample_frequency)
    #b, a = bilinear([1.0], [1.0, integration_time], fs=sample_frequency) # Bilinear: Analog to Digital filter.
    n = np.floor(integration_time * sample_frequency).astype(int)
    data = data[..., 0:n*(samples//n)]
    newshape = list(data.shape[0:-1])
    newshape.extend([-1, n])
    data = data.reshape(newshape)
    #data = data.reshape((-1, n)) # Divide in chunks over which to perform the integration.
    return lfilter(b, a, data)[...,n-1] / integration_time # Perform the integration. Select the final value of the integration.

def fast(data, fs):
    """Apply fast (F) time-weighting.
    
    :param data: Energetic quantity, e.g. :math:`p^2`.
    :param fs: Sample frequency.
    
    .. seealso:: :func:`integrate`
    
    """
    return integrate(data, fs, FAST)
    #return time_weighted_sound_level(data, fs, FAST)

def slow(data, fs):
    """Apply slow (S) time-weighting.
    
    :param data: Energetic quantity, e.g. :math:`p^2`.
    :param fs: Sample frequency.
    
    .. seealso:: :func:`integrate`
    
    """
    return integrate(data, fs, SLOW)
    #return time_weighted_sound_level(data, fs, SLOW)

def fast_level(data, fs):
    """Time-weighted (FAST) sound pressure level.
    
    :param data: Dynamic pressure.
    :param fs: Sample frequency.
    
    .. seealso:: :func:`time_weighted_sound_level`
    
    """
    return time_weighted_sound_level(data, fs, FAST)

def slow_level(data, fs):
    """Time-weighted (SLOW) sound pressure level.

    :param data: Dynamic pressure.
    :param fs: Sample frequency.
    
    .. seealso:: :func:`time_weighted_sound_level`
    
    """
    return time_weighted_sound_level(data, fs, SLOW)


#---- Annex E - Analytical expressions for frequency-weightings C, A, and Z.-#

_POLE_FREQUENCIES = {1: 20.60,
                     2: 107.7,
                     3: 737.9,
                     4: 12194.0,
                     }
"""Approximate values for pole frequencies f_1, f_2, f_3 and f_4.

See section E.4.1 of the standard.
"""

_NORMALIZATION_CONSTANTS = {'A': -2.000,
                            'C': -0.062,
                            }
"""Normalization constants :math:`C_{1000}` and :math:`A_{1000}`.

See section E.4.2 of the standard.
"""

def weighting_function_a(frequencies):
    """A-weighting function in decibel.
    
    :param frequencies: Vector of frequencies at which to evaluate the weighting.
    :returns: Vector with scaling factors.
    
    The weighting curve is
    
    .. math:: 20 \\log_{10}{\\frac{(f_4^2 * f^4)}{(f^2 + f_1^2) \sqrt{(f^2 + f_2^2)(f^2 + f_3^2)}(f^2 + f_4^2)}} - A_{1000}
    
    with :math:`A_{1000} = -2` dB.
    
    See equation E.6 of the standard.
    
    """
    f = np.asarray(frequencies)
    offset = _NORMALIZATION_CONSTANTS['A']
    f1, f2, f3, f4 = _POLE_FREQUENCIES.values()
    weighting = 20.0 * np.log10( (f4**2.0 * f**4.0) / ( (f**2.0+f1**2.0)*np.sqrt(f**2.0+f2**2.0)*np.sqrt(f**2.0+f3**2.0)*(f**2.0+f4**2.0) )) - offset
    return weighting


def weighting_function_c(frequencies):
    """C-weighting function in decibel.
    
    :param frequencies: Vector of frequencies at which to evaluate the weighting.
    :returns: Vector with scaling factors.
    
    The weighting curve is
    
    .. math:: 20 \\log_{10}{\\frac{(f_4^2 f^2)}{(f^2+f_1^2)(f^2+f_4^2)}} - C_{1000}
    
    with :math:`C_{1000} = -0.062` dB
    
    See equation E.1 of the standard.
    
    """
    f = np.asarray(frequencies)
    offset = _NORMALIZATION_CONSTANTS['C']
    f1, f2, f3, f4 = _POLE_FREQUENCIES.values()
    weighting = 20.0 * np.log10( (f4**2.0 * f**2.0)  / ( (f**2.0 + f1**2.0) * (f**2.0 + f4**2.0) )) - offset
    return weighting
    
    
def weighting_function_z(frequencies):
    """Z-weighting function in decibel.
    
    :param frequencies: Vector of frequencies at which to evaluate the weighting.
    :returns: Vector with scaling factors.
    
    """
    frequencies = np.asarray(frequencies)
    return np.zeros_like(frequencies)
    
    
WEIGHTING_FUNCTIONS = {'A': weighting_function_a,
                       'C': weighting_function_c,
                       'Z': weighting_function_z,
                       }
"""Dictionary with available weighting functions 'A', 'C' and 'Z'.
"""

def weighting_system_a():
    """A-weighting filter represented as polynomial transfer function.
    
    :returns: Tuple of `num` and `den`.
    
    See equation E.6 of the standard.
    
    """    
    f1 = _POLE_FREQUENCIES[1]
    f2 = _POLE_FREQUENCIES[2]
    f3 = _POLE_FREQUENCIES[3]
    f4 = _POLE_FREQUENCIES[4]
    offset = _NORMALIZATION_CONSTANTS['A']
    numerator = np.array([(2.0*np.pi*f4)**2.0 * (10**(-offset/20.0)), 0.0, 0.0, 0.0, 0.0])
    part1 = [1.0, 4.0*np.pi*f4, (2.0*np.pi*f4)**2.0]
    part2 = [1.0, 4.0*np.pi*f1, (2.0*np.pi*f1)**2.0]
    part3 = [1.0, 2.0*np.pi*f3]
    part4 = [1.0, 2.0*np.pi*f2]
    denomenator = np.convolve(np.convolve(np.convolve(part1, part2), part3), part4)
    return numerator, denomenator
    
    
def weighting_system_c():
    """C-weighting filter represented as polynomial transfer function.
    
    :returns: Tuple of `num` and `den`.
    
    See equation E.1 of the standard.
    
    """
    f1 = _POLE_FREQUENCIES[1]
    f4 = _POLE_FREQUENCIES[4]
    offset = _NORMALIZATION_CONSTANTS['C']
    numerator = np.array([(2.0*np.pi*f4)**2.0 * (10**(-offset/20.0)), 0.0, 0.0])
    part1 = [1.0, 4.0*np.pi*f4, (2.0*np.pi*f4)**2.0]
    part2 = [1.0, 4.0*np.pi*f1, (2.0*np.pi*f1)**2.0]
    denomenator = np.convolve(part1, part2)
    return numerator, denomenator
    
    
def weighting_system_z():
    """Z-weighting filter represented as polynomial transfer function.
    
    :returns: Tuple of `num` and `den`.
    
    Z-weighting is 0.0 dB for all frequencies and therefore corresponds to a 
    multiplication of 1.
    
    """
    numerator = [1]
    denomenator = [1]
    return numerator, denomenator


WEIGHTING_SYSTEMS = {'A': weighting_system_a,
                     'C': weighting_system_c,
                     'Z': weighting_system_z,
                     }
"""Weighting systems.
"""



"""
Quantities and units
====================

The Quantity module provides two classes to work with quantities and units.


.. inheritance-diagram:: acoustics.quantity

"""

from acoustics.standards.iso_tr_25417_2007 import REFERENCE_PRESSURE

quantities = {
    'pressure' : ('Pressure', 'pascal', True, 'p', '$p$', REFERENCE_PRESSURE)
    }
"""
Dictionary with quantities. Each quantity is stored as a tuple.
"""

units = {
    'meter' : ('meter', 'm', '$m$'),
    'pascal' : ('pascal', 'Pa', '$Pa$'),
    }
"""
Dictionary with units. Each unit is stored as a tuple.
"""

class Unit(object):
    """
    Unit of quantity.
    
    .. note:: Perhaps inherit from tuple or :class:`collections.namedTuple`?
    
    """
    def __init__(self, name, symbol, symbol_latex):
        
        self.name = name
        """
        Name of the unit.
        """
        self.symbol = symbol
        """
        Symbol of the unit.
        """
        
        self.symbol_latex
        """
        Symbol of the unit in LaTeX.
        """

    def __repr__(self):
        return "Unit({})".format(self.name)
    
    def __str__(self):
        return self.name


class Quantity(object):
    """
    Quantity.
    """
    def __init__(self, name, unit, dynamic, symbol=None, symbol_latex=None, reference=1.0):
        
        self.name = name
        """
        Name of the quantity.
        """
        
        self.symbol = symbol
        """
        Symbol of the quantity.
        """
        
        self.symbol_latex = symbol_latex
        """
        Symbol of the unit in LaTeX.
        """
        
        self.unit = unit
        """
        Unit. See :class:`Unit`.
        """
        
        self.dynamic = dynamic
        """
        Dynamic quantity (`True`) or energetic (`False`).
        """
        
        self.reference = reference
        """
        Reference value of the quantity.
        """
    
    def __repr__(self):
        return "Quantity({})".format(self.name)
    
    def __str__(self):
        return self.name

    @property
    def energetic(self):
        """
        Energetic quantity (`True`) or dynamic (`False`).
        """
        return not self.dynamic


def get_quantity(name):
    """
    Get quantity by name. Returns instance of :class:`Quantity`.
    
    :param name: Name of the quantity.
    
    """
    try:
        q = list(quantities[name])
    except KeyError:
        raise ValueError("Unknown quantity. Quantity is not yet specified.")
    try:
        u = units[name]
    except KeyError:
        raise RuntimeError("Unknown unit. Quantity has been specified but unit has not.")
    
    q[1] = Unit(*units[name])
    
    return Quantity(*q)


"""
Ambisonics
==========

"""

import numpy as np
from scipy.misc import factorial   

def acn(order=1):
    """Spherical harmonic degree `n` and order `m` for ambisonics order `order`.
    
    :param order: Ambisonics order.
    
    Yields tuples `(n, m)` where `n` is the degree and `m` the order. 
    
    Follows ACN.
    
    === == == ======
    ACN n  m  letter
    === == == ======
    0   0  0  W
    --- -- -- ------
    1   1  -1 Y
    --- -- -- ------
    2   1  0  Z
    --- -- -- ------
    3   1  +1 X
    === == == ======
    
    """
    for n in range(order+1):
        for m in range(-n, n+1):
            yield (n, m)

   
def sn3d(m, n):
    """SN3D or Schmidt semi-normalisation
    
    :param m: order `n`
    :param n: degree `m`
    
    http://en.wikipedia.org/wiki/Ambisonic_data_exchange_formats#SN3D
    
    """
    m = np.atleast_1d(m)
    n = np.atleast_1d(n)
    
    d = np.logical_not(m.astype(bool))
    out = np.sqrt( (2.0 - d)/(4.0*np.pi)  * factorial(n-np.abs(m)) / factorial(n+np.abs(m))  )
    return out


def n3d(m, n):
    """N3D or full three-D normalisation
    
    :param m: order `n`
    :param n: degree `m`
    
    http://en.wikipedia.org/wiki/Ambisonic_data_exchange_formats#N3D
    
    """
    n = np.atleast_1d(n)
    return sn3d(m, n) * np.sqrt(2*n+1)
    
    
__all__ = ['acn', 'sn3d', 'n3d']
    



"""
Building
========

The building module contains functions related to building acoustics.

"""
from __future__ import division

import numpy as np


def rw_curve(tl):
    """
    Calculate the curve of :math:`Rw` from a NumPy array `tl` with third
    octave data between 100 Hz and 3.15 kHz.
    
    :param tl: Transmission Loss
    
    """
    ref_curve = np.array([0, 3, 6, 9, 12, 15, 18, 19, 20, 21, 22, 23, 23, 23,
                          23, 23])
    residuals = 0
    while residuals > -32:
        ref_curve += 1
        diff = tl - ref_curve
        residuals = np.sum(np.clip(diff, np.min(diff), 0))
    ref_curve -= 1
    return ref_curve


def rw(tl):
    """
    Calculate :math:`R_W` from a NumPy array `tl` with third octave data
    between 100 Hz and 3.15 kHz.
    
    :param tl: Transmission Loss
    """
    return rw_curve(tl)[7]


def rw_c(tl):
    """
    Calculate :math:`R_W + C` from a NumPy array `tl` with third octave data
    between 100 Hz and 3.15 kHz.
    
    :param tl: Transmission Loss
    """
    k = np.array([-29, -26, -23, -21, -19, -17, -15, -13, -12, -11, -10, -9,
                  -9, -9, -9, -9])
    a = -10 * np.log10(np.sum(10**((k - tl)/10)))
    return a


def rw_ctr(tl):
    """
    Calculate :math:`R_W + C_{tr}` from a NumPy array `tl` with third octave
    data between 100 Hz and 3.15 kHz.
    
    :param tl: Transmission Loss
    """
    k_tr = np.array([-20, -20, -18, -16, -15, -14, -13, -12, -11, -9, -8, -9,
                     -10, -11, -13, -15])
    a_tr = -10 * np.log10(np.sum(10**((k_tr - tl)/10)))
    return a_tr


def stc_curve(tl):
    """
    Calculate the Sound Transmission Class (STC) curve from a NumPy array `tl`
    with third octave data between 125 Hz and 4 kHz.
    
    :param tl: Transmission Loss
    """
    ref_curve = np.array([0, 3, 6, 9, 12, 15, 16, 17, 18, 19, 20, 20, 20,
                            20, 20, 20])
    top_curve = ref_curve
    res_sum = 0
    while True:
        diff = tl - top_curve
        residuals = np.clip(diff, np.min(diff), 0)
        res_sum = np.sum(residuals)
        if res_sum < -32:
            if np.any(residuals > -8):
                top_curve -= 1
                break
        top_curve += 1
    return top_curve


def stc(tl):
    """
    Calculate the Sound Transmission Class (STC) from a NumPy array `tl` with
    third octave data between 125 Hz and 4 kHz.
    
    :param tl: Transmission Loss
    """
    return stc_curve(tl)[6]


def mass_law(freq, vol_density, thickness, theta=0, c=343, rho0=1.225):
    """ Calculate transmission loss according to mass law.
    
    :param freq: Frequency of interest in Hz.
    :type freq: `float` or `NumPy array`
    :param vol_density: Volumetric density of material in [kg/m^3].
    :type vol_density: `float`
    :param thickness: Thickness of wall.
    :type thickness: `float`
    :param theta: Angle of incidence in degrees. Default value is `0` (normal incidence).
    :type theta: `float`
    :param c: Speed of sound in [m/s].
    :type c: `float`
    :param rho0: Density of air in kg/m^3.
    :type rho0: `float`
    """
    rad_freq = 2.0 * np.pi * freq
    surface_density = vol_density * thickness
    theta_rad = np.deg2rad(theta)
    a = rad_freq * surface_density * np.cos(theta_rad) / (2 * rho0 * c)
    tl_theta = 10 * np.log10(1 + a**2)
    return tl_theta


__all__ = ['rw_curve', 'rw', 'rw_c', 'rw_ctr', 'stc_curve', 'stc', 'mass_law']



"""
IEC 61260-1:2014
================

IEC 61260-1:2014 specifies performance requirements for analogue, sampled-data, 
and digital implementations of band-pass filters. 


Frequency functions
**************************************

.. autofunction:: acoustics.standards.iec_61260_1_2014.exact_center_frequency
.. autofunction:: acoustics.standards.iec_61260_1_2014.lower_frequency
.. autofunction:: acoustics.standards.iec_61260_1_2014.upper_frequency
.. autofunction:: acoustics.standards.iec_61260_1_2014.index_of_frequency


Nominal center frequencies
**************************

.. autoattribute:: acoustics.standards.iec_61260_1_2014.NOMINAL_OCTAVE_CENTER_FREQUENCIES
.. autoattribute:: acoustics.standards.iec_61260_1_2014.NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES

.. autoattribute:: acoustics.standards.iec_61260_1_2014.REFERENCE_FREQUENCY
.. autoattribute:: acoustics.standards.iec_61260_1_2014.OCTAVE_FREQUENCY_RATIO

"""
from __future__ import division

import acoustics
import numpy as np

NOMINAL_OCTAVE_CENTER_FREQUENCIES = np.array([31.5, 63.0, 125.0, 250.0, 
                                              500.0, 1000.0, 2000.0, 
                                              4000.0, 8000.0, 16000.0])
"""Nominal octave center frequencies.
"""

NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES = np.array([25.0, 31.5, 40.0, 50.0, 63.0, 80.0, 100.0, 125.0, 160.0, 
                                                    200.0, 250.0, 315.0, 400.0, 500.0, 630.0, 800.0, 1000.0, 
                                                    1250.0, 1600.0, 2000.0, 2500.0, 3150.0, 4000.0, 5000.0, 
                                                    6300.0, 8000.0, 10000.0, 12500.0, 16000.0, 20000.0
                                                    ])
"""Nominal third-octave center frequencies in the audio range.
"""



REFERENCE_FREQUENCY = 1000.0
"""Reference frequency.
"""

OCTAVE_FREQUENCY_RATIO = 10.0**(3.0/10.0)
"""Octave frequency ratio :math:`G`.

See equation 1.
"""

def exact_center_frequency(x, fraction=1, ref=REFERENCE_FREQUENCY, G=OCTAVE_FREQUENCY_RATIO):
    """
    Center frequencies :math:`f_m` for band indices :math:`x`. See equation 2 and 3.
    
    :param x: Band index :math:`x`.
    :param ref: Reference center frequency :math:`f_r`.
    :param fraction: Bandwidth designator :math`b`. For example, for 1/3-octave filter b=3.
    :param G: Octave frequency ratio :math:`G`.
    
    The center frequencies are given by
    
    .. math:: f_m = f_r G^{x/b}
    
    In case the bandwidth designator :math:`b` is an even number, the center frequencies are given by
    
    .. math:: f_m = f_r G^{(2x+1)/2b}
    
    See equation 2 and 3 of the standard.
    """
    #if fraction%2: # Uneven
        #return ref * G**(x / fraction)
    #else: # Even
        #return ref * G**((2.*x+1.0) / (2.0*fraction))
    fraction = np.asarray(fraction)
    uneven = (fraction%2).astype('bool')
    return ref * G**((2.0*x+1.0) / (2.0*fraction)) * np.logical_not(uneven) + uneven * ref * G**(x / fraction)



def lower_frequency(center, fraction=1, G=OCTAVE_FREQUENCY_RATIO):
    """
    Lower band-edge frequencies. See equation 4.
    
    :param center: Center frequencies :math:`f_m`.
    :param fraction: Bandwidth designator :math:`b`.
    :param G: Octave frequency ratio :math:`G`.
    
    The lower band-edge frequencies are given by
    
    .. math:: f_1 = f_m G^{-1/2b}
    
    See equation 4 of the standard.
    
    """
    return center * G**(-1.0 / (2.0*fraction))


def upper_frequency(center, fraction=1, G=OCTAVE_FREQUENCY_RATIO):
    """
    Upper band-edge frequencies. See equation 5.
    
    :param center: Center frequencies :math:`f_m`.
    :param fraction: Bandwidth designator :math:`b`.
    :param G: Octave frequency ratio :math:`G`.
    
    The upper band-edge frequencies are given by
    
    .. math:: f_2 = f_m G^{+1/2b}
    
    See equation 5 of the standard.
    
    """
    return center * G**(+1.0 / (2.0*fraction))


def index_of_frequency(frequency, fraction=1, ref=REFERENCE_FREQUENCY, G=OCTAVE_FREQUENCY_RATIO):
    """Determine the band index `x` from a given frequency.
    
    :param frequency: Frequencies :math:`f`.
    :param fraction: Bandwidth designator :math:`b`.
    :param ref: Reference frequency.
    :param G: Octave frequency ratio :math:`G`.
    
    The index of the center frequency is given by
    
    .. math:: x = round{b \\frac{\log{f/f_{ref} }}{\log{G} }}        
    
    .. note:: This equation is not part of the standard. However, it follows from :func:`exact_center_frequency`.
    
    """
    #if fraction%2: # Uneven
        #return np.round(fraction * np.log(frequency/ref) / np.log(G)).astype('int16')
    #else: # Even
        #return np.round((2.0*fraction * np.log(frequency/ref) / np.log(G) - 1.0)) / 2.0
    fraction = np.asarray(fraction)
    uneven = (fraction%2).astype('bool')
    return (np.round((2.0*fraction * np.log(frequency/ref) / np.log(G) - 1.0)) / 2.0 * np.logical_not(uneven) + uneven * np.round(fraction * np.log(frequency/ref) / np.log(G)).astype('int16')).astype('int16')


def _nominal_center_frequency(center, fraction):
    """Nominal frequency according to standard.
    
    :param center: Exact mid-frequency to be rounded.
    :param fraction: Bandwidth designator or fraction.
    """
    def _roundn(x, n):
        return round(x, -int(np.floor(np.sign(x) * np.log10(abs(x)))) + n)
    
    b = fraction
    x = center
    
    # Section E.1: 1/1-octaves
    if b == 1:
        n = index_of_frequency(x, b)
        if -6 <= n < 5: # Correspond to indices when n=0 corresponds to 1000 Hz
            return acoustics.standards.iec_61672_1_2013.NOMINAL_OCTAVE_CENTER_FREQUENCIES[n+6]
        elif n >= 5:
            return 2.0 * _nominal_center_frequency(exact_center_frequency(n-1, b), b) # WARNING: Unclear in standard!
        else:
            return 1./2.0 * _nominal_center_frequency(exact_center_frequency(n+1, b), b) # WARNING: Unclear in standard!
        
    # Section E.2: 1/2-octaves
    elif b == 2:
        return _roundn(x, 2)

    # Section E.1: 1/3-octaves
    elif b == 3:
        n = index_of_frequency(x, b)
        
        if -20 <= n < 14: # Correspond to indices when n=0 corresponds to 1000 Hz
            return acoustics.standards.iec_61672_1_2013.NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES[n+20]
        elif n >= 14:
            return 10.*_nominal_center_frequency(exact_center_frequency(n-10, b), b) # WARNING: Unclear in standard!
        else:
            return 1./10.*_nominal_center_frequency(exact_center_frequency(n+10, b), b) # WARNING: Unclear in standard!
    
    # Section E3.3: 1/4 to 1/24-octaves, inclusive
    elif 4 <= b <= 24:
        msd = x // 10.0**np.floor(np.log10(x))
        if msd < 5:
            return _roundn(x, 2) # E3.2
        else:
            return _roundn(x, 1) # E3.3
   
    # Section E3.5: > 1/24-octaves
    elif b > 24:
        raise NotImplementedError("b > 24 is not implemented")
    else:
        raise ValueError("Wrong value for b")

nominal_center_frequency = np.vectorize(_nominal_center_frequency)
"""Nominal center frequency.

:param center: Exact center frequency.
:param fraction: Band designator or fraction.

"""
    #pass
    




"""
ISO 1683:2015
=============
"""
REFERENCE_VALUES = {
  "gas" : {
    "pressure"  : 2e-5,         # Pa
    "exposure"  : (2e-5)**2,    # Pa^2 s
    "power"     : 1e-12,        # W
    "energy"    : 1e-12,        # J
    "intensity" : 1e-12,        # W/m^2
    },
  "liquid" : {
      "pressure"    : 1e-5,     # Pa
      "exposure"    : 1e-5,     # Pa^2 s
      "power"       : 1e-12,    # W
      "energy"      : 1e-12,    # J
      "intensity"   : 1e-12,    # W/m^2
      "particle_displacement"   :   1e-12,  # m
      "particle_velocity"       :   1e-9,   # m/s
      "particle_acceleration"   :   1e-6,   # m/s^2
      "distance"                :   1,      # m
      },
  "vibration" : {
      "displacement"    :   1e-12,  # m
      "velocity"        :   1e-9,   # m/s
      "acceleration"    :   1e-6,   # m/s^2
      "force"           :   1e-6,   # N}
      }
  }



"""
Descriptors
============

The descriptors module offers all kinds of acoustics related descriptors.

.. toctree::
    :maxdepth: 2

Descriptors from ISO/TR 25417:2007
**********************************

Descriptors from :mod:`acoustics.standards.iso_tr_25417_2007`.

.. autoattribute:: acoustics.descriptors.REFERENCE_PRESSURE 
.. autofunction:: acoustics.descriptors.sound_pressure_level 
.. autofunction:: acoustics.descriptors.equivalent_sound_pressure_level
.. autofunction:: acoustics.descriptors.peak_sound_pressure
.. autofunction:: acoustics.descriptors.peak_sound_pressure_level
.. autoattribute:: acoustics.descriptors.REFERENCE_SOUND_EXPOSURE
.. autofunction:: acoustics.descriptors.sound_exposure
.. autofunction:: acoustics.descriptors.sound_exposure_level
.. autoattribute:: acoustics.descriptors.REFERENCE_POWER
.. autofunction:: acoustics.descriptors.sound_power_level
.. autofunction:: acoustics.descriptors.sound_energy
.. autofunction:: acoustics.descriptors.sound_energy_level
.. autoattribute:: acoustics.descriptors.REFERENCE_ENERGY
.. autofunction:: acoustics.descriptors.sound_intensity
.. autofunction:: acoustics.descriptors.time_averaged_sound_intensity
.. autoattribute:: acoustics.descriptors.REFERENCE_INTENSITY
.. autofunction:: acoustics.descriptors.time_averaged_sound_intensity_level
.. autofunction:: acoustics.descriptors.normal_time_averaged_sound_intensity
.. autofunction:: acoustics.descriptors.normal_time_averaged_sound_intensity_level


Other descriptors
*****************

"""

from __future__ import division

import numpy as np

from acoustics.standards.iso_tr_25417_2007 import (REFERENCE_PRESSURE, 
                                                   sound_pressure_level, 
                                                   equivalent_sound_pressure_level,
                                                   peak_sound_pressure,
                                                   peak_sound_pressure_level,
                                                   REFERENCE_SOUND_EXPOSURE,
                                                   sound_exposure,
                                                   sound_exposure_level,
                                                   REFERENCE_POWER,
                                                   sound_power_level,
                                                   sound_energy,
                                                   REFERENCE_ENERGY,
                                                   sound_energy_level,
                                                   sound_intensity,
                                                   time_averaged_sound_intensity,
                                                   REFERENCE_INTENSITY,
                                                   time_averaged_sound_intensity_level,
                                                   normal_time_averaged_sound_intensity,
                                                   normal_time_averaged_sound_intensity_level,
                                                   )


from acoustics.standards.iso_1996_1_2003 import composite_rating_level

def _leq(levels, time):
    levels = np.asarray(levels)
    return 10.0 * np.log10((1.0/time) * np.sum(10.0**(levels/10.0)))


def leq(levels, int_time=1.0):
    """
    Equivalent level :math:`L_{eq}`.
    
    :param levels: Levels as function of time.
    :param int_time: Integration time. Default value is 1.0 second.
    :returns: Equivalent level L_{eq}.
    
    Sum of levels in dB.
    """
    levels = np.asarray(levels)
    time = levels.size * int_time
    return _leq(levels, time)


def sel(levels):
    """
    Sound Exposure Level from ``levels`` (NumPy array).
    """
    levels = np.asarray(levels)
    return _leq(levels, 1.0)


def lw(W, Wref=1.0e-12):
    """
    Sound power level :math:`L_{w}` for sound power :math:`W` and reference power :math:`W_{ref}`.
    
    :param W: Sound power :math:`W`.
    :param Wref: Reference power :math:`W_{ref}`. Default value is :math:`10^{12}` watt.
    """
    W = np.asarray(W)
    return 10.0 * np.log10(W/Wref)


def lden(lday, levening, lnight, hours=(12.0, 4.0, 8.0), adjustment=(0.0, 5.0, 10.0)):
    """
    Calculate :math:`L_{den}` from :math:`L_{day}`, :math:`L_{evening}` and :math:`L_{night}`.
    
    :param lday: Equivalent level during day period :math:`L_{day}`.
    :param levening: Equivalent level during evening period :math:`L_{evening}`.
    :param lnight: Equivalent level during night period :math:`L_{night}`.
    :param hours: Hours per period.
    :param adjustment: Correction factor per period.
    :returns: :math:`L_{den}`
    
    .. seealso:: :func:`acoustics.standards.iso_1996_1_2003.composite_rating_level`
    """
    lday = np.asarray(lday)
    levening = np.asarray(levening)
    lnight = np.asarray(lnight)
    return composite_rating_level(np.vstack((lday, levening, lnight)).T, hours, adjustment)


def ldn(lday, lnight, hours=(15.0, 9.0), adjustment=(0.0, 10.0)):
    """
    Calculate :math:`L_{dn}` from :math:`L_{day}` and :math:`L_{night}`.
    
    :param lday: Equivalent level during day period :math:`L_{day}`.
    :param lnight: Equivalent level during night period :math:`L_{night}`.
    :param hours: Hours per period.
    :param adjustment: Correction factor per period.
    :returns: :math:`L_{dn}`
    
    .. seealso:: :func:`acoustics.standards.iso_1996_1_2003.composite_rating_level`
    """
    lday = np.asarray(lday)
    lnight = np.asarray(lnight)
    return composite_rating_level(np.vstack((lday, lnight)).T, hours, adjustment)



"""
Octave
======

Module for working with octaves.

The following is an example on how to use :class:`acoustics.octave.Octave`.

.. literalinclude:: ../examples/octave.py

"""
from __future__ import division

import numpy as np
import acoustics

#REFERENCE = 1000.0
#"""
#Reference frequency.
#"""

from acoustics.standards.iec_61260_1_2014 import index_of_frequency
from acoustics.standards.iec_61260_1_2014 import REFERENCE_FREQUENCY as REFERENCE


def exact_center_frequency(frequency=None, fraction=1, n=None, ref=REFERENCE):
    """Exact center frequency.
    
    :param frequency: Frequency within the band.
    :param fraction: Band designator.
    :param n: Index of band.
    :param ref: Reference frequency.
    :return: Exact center frequency for the given frequency or band index.
    
    .. seealso:: :func:`acoustics.standards.iec_61260_1_2014.exact_center_frequency`
    .. seealso:: :func:`acoustics.standards.iec_61260_1_2014.index_of_frequency`
    
    """
    if frequency is not None:
        n = acoustics.standards.iec_61260_1_2014.index_of_frequency(frequency, fraction=fraction, ref=ref)
    return acoustics.standards.iec_61260_1_2014.exact_center_frequency(n, fraction=fraction, ref=ref)


def nominal_center_frequency(frequency=None, fraction=1, n=None):
    """Nominal center frequency.
    
    :param frequency: Frequency within the band.
    :param fraction: Band designator.
    :param n: Index of band.
    :returns: The nominal center frequency for the given frequency or band index.
    
    .. seealso:: :func:`acoustics.standards.iec_61260_1_2014.exact_center_frequency`
    .. seealso:: :func:`acoustics.standards.iec_61260_1_2014.nominal_center_frequency`
    
    .. note:: Contrary to the other functions this function silently assumes 1000 Hz reference frequency.
    
    """
    center = exact_center_frequency(frequency, fraction, n)
    return acoustics.standards.iec_61260_1_2014.nominal_center_frequency(center, fraction)
    

def lower_frequency(frequency=None, fraction=1, n=None, ref=REFERENCE):
    """Lower band-edge frequency.
    
    :param frequency: Frequency within the band.
    :param fraction: Band designator.
    :param n: Index of band.
    :param ref: Reference frequency.
    :returns: Lower band-edge frequency for the given frequency or band index.
    
    .. seealso:: :func:`acoustics.standards.iec_61260_1_2014.exact_center_frequency`
    .. seealso:: :func:`acoustics.standards.iec_61260_1_2014.lower_frequency`
    
    """
    center = exact_center_frequency(frequency, fraction, n, ref=ref)
    return acoustics.standards.iec_61260_1_2014.lower_frequency(center, fraction)
   
   
def upper_frequency(frequency=None, fraction=1, n=None, ref=REFERENCE):
    """Upper band-edge frequency.
    
    :param frequency: Frequency within the band.
    :param fraction: Band designator.
    :param n: Index of band.
    :param ref: Reference frequency.
    :returns: Upper band-edge frequency for the given frequency or band index.
    
    .. seealso:: :func:`acoustics.standards.iec_61260_1_2014.exact_center_frequency`
    .. seealso:: :func:`acoustics.standards.iec_61260_1_2014.upper_frequency`
    
    """
    center = exact_center_frequency(frequency, fraction, n, ref=ref)
    return acoustics.standards.iec_61260_1_2014.upper_frequency(center, fraction)


#-- things below will be deprecated?---#

frequency_of_band = acoustics.standards.iec_61260_1_2014.exact_center_frequency
band_of_frequency = index_of_frequency


class Octave(object):
    """
    Class to calculate octave center frequencies.
    """

    def __init__(self, fraction=1, interval=None, fmin=None, fmax=None, unique=False, reference=REFERENCE):
        
        self.reference = reference
        """
        Reference center frequency :math:`f_{c,0}`.
        """
        
        self.fraction = fraction
        """
        Fraction of octave.
        """
        
        if (interval is not None) and (fmin is not None or fmax is not None):
            raise AttributeError
        
        self._interval = np.asarray(interval)
        """Interval"""
        
        self._fmin = fmin
        """Minimum frequency of a range."""
        
        self._fmax = fmax
        """Maximum frequency of a range."""
        
        self.unique = unique
        """Whether or not to calculate the requested values for every value of ``interval``."""
        
        
    @property
    def fmin(self):
        """Minimum frequency of an interval."""
        if self._fmin is not None:
            return self._fmin
        elif self._interval is not None:
            return self.interval.min()
    
    @fmin.setter
    def fmin(self, x):
        if self.interval is not None:
            pass    # Warning, remove interval first.
        else:
            self._fmin = x
    
    @property
    def fmax(self):
        """Maximum frequency of an interval."""
        if self._fmax is not None:
            return self._fmax
        elif self._interval is not None:
            return self.interval.max()
    
    @fmax.setter
    def fmax(self, x):
        if self.interval is not None:
            pass
        else:
            self._fmax = x

    @property
    def interval(self):
        """Interval."""
        return self._interval
    
    @interval.setter
    def interval(self, x):
        if self._fmin or self._fmax:
            pass
        else:
            self._interval = np.asarray(x)
        
    def _n(self, f):
        """
        Calculate the band ``n`` from a given frequency.
        
        :param f: Frequency
        
        See also :func:`band_of_frequency`.
        """
        return band_of_frequency(f, fraction=self.fraction, ref=self.reference)
        
    def _fc(self, n):
        """
        Calculate center frequency of band ``n``.
        
        :param n: band ``n`.
        
        See also :func:`frequency_of_band`.
        """
        return frequency_of_band(n, fraction=self.fraction, ref=self.reference)
    
    @property
    def n(self):
        """
        Return band ``n`` for a given frequency.
        """
        if self.interval is not None and self.unique:
            return self._n(self.interval)
        else:
            return np.arange(self._n(self.fmin), self._n(self.fmax)+1)
    
    @property
    def center(self):
        """
        Return center frequencies :math:`f_c`.
        
        .. math::  f_c = f_{ref} \cdot 2^{n/N} \\cdot 10^{\\frac{3}{10N}}
        
        """
        n = self.n
        return self._fc(n)
     
    @property
    def bandwidth(self):
        """
        Bandwidth of bands.
        
        .. math:: B = f_u - f_l
        
        """
        return self.upper - self.lower
    
    @property
    def lower(self):
        """
        Lower frequency limits of bands.
        
        .. math:: f_l = f_c \cdot 2^{\\frac{-1}{2N}}
        
        See also :func:`lower_frequency`.
        """
        return lower_frequency(self.center, self.fraction)
    
    @property
    def upper(self):
        """
        Upper frequency limits of bands.
        
        .. math:: f_u = f_c \cdot 2^{\\frac{+1}{2N}}
        
        See also :func:`upper_frequency`.
        """
        return upper_frequency(self.center, self.fraction)
        

__all__ = ['exact_center_frequency', 'nominal_center_frequency', 
           'lower_frequency', 'upper_frequency', 'index_of_frequency',
           'Octave', 'frequency_of_band', 'band_of_frequency', # These three will be deprecated?
           ]

###def center_frequency_octave(frequencies, fraction=1):
    ###"""
    ###Calculate the center frequencies :math:`f_c` of the octaves that (partially) cover ``frequencies``.
    
    ###:param frequencies: An iterable containing frequencies.
    ###:param order: An integer indicating the octave order. E.g., for 1/3-octaves use ``order=3``
    
    ###Center frequencies are calculated using:
    
    ###.. math::  f_c = 1000 \cdot 2^{n/N} \\cdot 10^{\\frac{3}{10N}}
    
    ###"""
    ###n = lambda fc, N: np.log2(fc/1000.0) - 1.0/N    # To calculate the nth octave from a given frequency.
    
    ###n_min = np.floor(n(np.min(frequencies), order))
    ###n_max = np.ceil(n(np.max(frequencies), order))
    ###n = np.arange(n_min, n_max+1)
    ###fc = 1000.0 * 10.0**(3.0/order/10.0) * 2.0**(n)
    ###return fc



"""
Generator
=========

The generator module provides signal generators.

The following functions calculate ``N`` samples and return an array containing the samples.

For indefinitely long iteration over the samples, consider using the output of these functions in :func:`itertools.cycle`.

Noise
*****

Different types of noise are available. The following table lists the color 
of noise and how the power and power density change per octave.

====== ===== =============
Color  Power Power density
====== ===== =============
White  +3 dB  0 dB
Pink    0 dB -3 dB
Blue   +6 dB +3 dB
Brown  -3 dB -6 dB
Violet +9 dB +6 dB
====== ===== =============

The colored noise is created by generating pseudo-random numbers using
:func:`np.random.randn` and then multiplying these with a curve tyical for the color.
Afterwards, an inverse DFT is performed using :func:`np.fft.irfft`.
Finally, the noise is normalized using :func:`acoustics.signal.normalize`.

All colors
----------

.. autofunction:: noise
.. autofunction:: noise_generator

Per color
---------

.. autofunction:: white
.. autofunction:: pink
.. autofunction:: blue
.. autofunction:: brown
.. autofunction:: violet


Other
*****

.. autofunction:: heaviside

For related functions, check :mod:`scipy.signal`.


"""

import numpy as np
import random
import itertools
#import scipy.signal.sawtooth

try:
    from pyfftw.interfaces.numpy_fft import rfft, irfft       # Performs much better than numpy's fftpack
except ImportError:                                    # Use monkey-patching np.fft perhaps instead?
    from numpy.fft import rfft, irfft

from .signal import normalize


def noise(N, color='white'):
    """Noise generator.
    
    :param N: Amount of samples.
    :param color: Color of noise.
    
    """
    try:
        return _noise_generators[color](N)
    except KeyError:
        raise ValueError("Incorrect color.")


def white(N):
    """
    White noise.
    
    :param N: Amount of samples.
    
    White noise has a constant power density. It's narrowband spectrum is therefore flat.
    The power in white noise will increase by a factor of two for each octave band, 
    and therefore increases with 3 dB per octave.
    """
    return np.random.randn(N)


def pink(N):
    """
    Pink noise. 
    
    :param N: Amount of samples.
    
    Pink noise has equal power in bands that are proportionally wide.
    Power density decreases with 3 dB per octave.
    
    """
    # This method uses the filter with the following coefficients.
    #b = np.array([0.049922035, -0.095993537, 0.050612699, -0.004408786])
    #a = np.array([1, -2.494956002, 2.017265875, -0.522189400])
    #return lfilter(B, A, np.random.randn(N))
    # Another way would be using the FFT
    #x = np.random.randn(N)
    #X = rfft(x) / N  
    uneven = N%2
    X = np.random.randn(N//2+1+uneven) + 1j * np.random.randn(N//2+1+uneven)
    S = np.sqrt(np.arange(len(X))+1.) # +1 to avoid divide by zero
    y = (irfft(X/S)).real
    if uneven:
        y = y[:-1]
    return normalize(y)


def blue(N):
    """
    Blue noise. 
    
    :param N: Amount of samples.
    
    Power increases with 6 dB per octave.
    Power density increases with 3 dB per octave. 
    
    """
    uneven = N%2
    X = np.random.randn(N//2+1+uneven) + 1j * np.random.randn(N//2+1+uneven)
    S = np.sqrt(np.arange(len(X)))# Filter
    y = (irfft(X*S)).real
    if uneven:
        y = y[:-1]
    return normalize(y)


def brown(N):
    """
    Violet noise.
    
    :param N: Amount of samples.
    
    Power decreases with -3 dB per octave.
    Power density decreases with 6 dB per octave. 

    """
    uneven = N%2
    X = np.random.randn(N//2+1+uneven) + 1j * np.random.randn(N//2+1+uneven)
    S = (np.arange(len(X))+1)# Filter
    y = (irfft(X/S)).real
    if uneven:
        y = y[:-1]
    return normalize(y)


def violet(N):
    """
    Violet noise. Power increases with 6 dB per octave. 
    
    :param N: Amount of samples.
    
    Power increases with +9 dB per octave.
    Power density increases with +6 dB per octave. 
    
    """
    uneven = N%2
    X = np.random.randn(N//2+1+uneven) + 1j * np.random.randn(N//2+1+uneven)
    S = (np.arange(len(X)))# Filter
    y = (irfft(X*S)).real
    if uneven:
        y = y[:-1]
    return normalize(y)


_noise_generators = {
    'white'  : white,
    'pink'   : pink,
    'blue'   : blue,
    'brown'  : brown,
    'violet' : violet,
    }


def noise_generator(N=44100, color='white'):
    """Noise generator. 

    :param N: Amount of unique samples to generate.
    :param color: Color of noise.
     
    Generate `N` amount of unique samples and cycle over these samples.
    
    """
    #yield from itertools.cycle(noise(N, color)) # Python 3.3
    for sample in itertools.cycle(noise(N, color)):
        yield sample
    

def heaviside(N):
    """Heaviside.
    
    Returns the value 0 for `x < 0`, 1 for `x > 0`, and 1/2 for `x = 0`.
    """
    return 0.5 * (np.sign(N) + 1)
    

__all__ = ['noise', 'white', 'pink', 'blue', 'brown', 'violet',
           'noise_generator', 'heaviside']



"""
Signal
======

The signal module constains all kinds of signal processing related functions.

.. inheritance-diagram:: acoustics.signal


Filtering
*********

.. autoclass:: Filterbank
.. autofunction:: bandpass_filter
.. autofunction:: octave_filter
.. autofunction:: bandpass
.. autofunction:: lowpass
.. autofunction:: highpass
.. autofunction:: octavepass
.. autofunction:: convolve

Windowing
*********

.. autofunction:: window_scaling_factor
.. autofunction:: apply_window

Spectra
*******

Different types of spectra exist.

.. autofunction:: amplitude_spectrum
.. autofunction:: auto_spectrum
.. autofunction:: power_spectrum
.. autofunction:: density_spectrum
.. autofunction:: angle_spectrum
.. autofunction:: phase_spectrum

Frequency bands
***************

.. autoclass:: Band
.. autoclass:: Frequencies
.. autoclass:: EqualBand
.. autoclass:: OctaveBand

.. autofunction:: integrate_bands
.. autofunction:: octaves
.. autofunction:: third_octaves


Hilbert transform
*****************

.. autofunction:: amplitude_envelope
.. autofunction:: instantaneous_phase
.. autofunction:: instantaneous_frequency


Conversion
**********

.. autofunction:: decibel_to_neper
.. autofunction:: neper_to_decibel


Other
*****

.. autofunction:: isolate
.. autofunction:: zero_crossings
.. autofunction:: rms
.. autofunction:: ms
.. autofunction:: normalize
.. autofunction:: ir2fr
.. autofunction:: wvd

"""
from __future__ import division

import matplotlib.pyplot as plt
import numpy as np
from scipy.sparse import spdiags
from scipy.signal import butter, lfilter, freqz, filtfilt, sosfilt

import acoustics.octave
#from acoustics.octave import REFERENCE

import acoustics.bands
from scipy.signal import hilbert
from acoustics.standards.iso_tr_25417_2007 import REFERENCE_PRESSURE
from acoustics.standards.iec_61672_1_2013 import (NOMINAL_OCTAVE_CENTER_FREQUENCIES,
                                                  NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES)

try:
    from pyfftw.interfaces.numpy_fft import rfft
except ImportError:
    from numpy.fft import rfft

def bandpass_filter(lowcut, highcut, fs, order=8, output='sos'):
    """Band-pass filter.

    :param lowcut: Lower cut-off frequency
    :param highcut: Upper cut-off frequency
    :param fs: Sample frequency
    :param order: Filter order
    :param output: Output type. {'ba', 'zpk', 'sos'}. Default is 'sos'. See also :func:`scipy.signal.butter`.
    :returns: Returned value depends on `output`.

    A Butterworth filter is used.

    .. seealso:: :func:`scipy.signal.butter`.

    """
    nyq = 0.5 * fs
    low = lowcut / nyq
    high = highcut / nyq
    output = butter(order/2, [low, high], btype='band', output=output)
    return output


def bandpass(signal, lowcut, highcut, fs, order=8, zero_phase=False):
    """Filter signal with band-pass filter.

    :param signal: Signal
    :param lowcut: Lower cut-off frequency
    :param highcut: Upper cut-off frequency
    :param fs: Sample frequency
    :param order: Filter order
    :param zero_phase: Prevent phase error by filtering in both directions (filtfilt)

    A Butterworth filter is used. Filtering is done with second-order sections.

    .. seealso:: :func:`bandpass_filter` for the filter that is used.

    """
    sos = bandpass_filter(lowcut, highcut, fs, order, output='sos')
    if zero_phase:
        return _sosfiltfilt(sos, signal)
    else:
        return sosfilt(sos, signal)


def bandstop(signal, lowcut, highcut, fs, order=8, zero_phase=False):
    """Filter signal with band-stop filter.

    :param signal: Signal
    :param lowcut: Lower cut-off frequency
    :param highcut: Upper cut-off frequency
    :param fs: Sample frequency
    :param order: Filter order
    :param zero_phase: Prevent phase error by filtering in both directions (filtfilt)

    """
    return lowpass(signal, lowcut, fs, order=(order//2), zero_phase=zero_phase) + highpass(signal, highcut, fs, order=(order//2), zero_phase=zero_phase)


def lowpass(signal, cutoff, fs, order=4, zero_phase=False):
    """Filter signal with low-pass filter.

    :param signal: Signal
    :param fs: Sample frequency
    :param cutoff: Cut-off frequency
    :param order: Filter order
    :param zero_phase: Prevent phase error by filtering in both directions (filtfilt)

    A Butterworth filter is used. Filtering is done with second-order sections.

    .. seealso:: :func:`scipy.signal.butter`.

    """
    sos = butter(order, cutoff/(fs/2.0), btype='low', output='sos')
    if zero_phase:
        return _sosfiltfilt(sos, signal)
    else:
        return sosfilt(sos, signal)


def highpass(signal, cutoff, fs, order=4, zero_phase=False):
    """Filter signal with low-pass filter.

    :param signal: Signal
    :param fs: Sample frequency
    :param cutoff: Cut-off frequency
    :param order: Filter order
    :param zero_phase: Prevent phase error by filtering in both directions (filtfilt)

    A Butterworth filter is used. Filtering is done with second-order sections.

    .. seealso:: :func:`scipy.signal.butter`.

    """
    sos = butter(order, cutoff/(fs/2.0), btype='high', output='sos')
    if zero_phase:
        return _sosfiltfilt(sos, signal)
    else:
        return sosfilt(sos, signal)


def octave_filter(center, fs, fraction, order=8, output='sos'):
    """Fractional-octave band-pass filter.

    :param center: Centerfrequency of fractional-octave band.
    :param fs: Sample frequency
    :param fraction: Fraction of fractional-octave band.
    :param order: Filter order
    :param output: Output type. {'ba', 'zpk', 'sos'}. Default is 'sos'. See also :func:`scipy.signal.butter`.

    A Butterworth filter is used.

    .. seealso:: :func:`bandpass_filter`

    """
    ob = OctaveBand(center=center, fraction=fraction)
    return bandpass_filter(ob.lower[0], ob.upper[0], fs, order, output=output)


def octavepass(signal, center, fs, fraction, order=8, zero_phase=True):
    """Filter signal with fractional-octave bandpass filter.

    :param signal: Signal
    :param center: Centerfrequency of fractional-octave band.
    :param fs: Sample frequency
    :param fraction: Fraction of fractional-octave band.
    :param order: Filter order
    :param zero_phase: Prevent phase error by filtering in both directions (filtfilt)

    A Butterworth filter is used. Filtering is done with second-order sections.

    .. seealso:: :func:`octave_filter`

    """
    sos = octave_filter(center, fs, fraction, order)
    if zero_phase:
        return _sosfiltfilt(sos, signal)
    else:
        return sosfilt(sos, signal)


def convolve(signal, ltv, mode='full'):
    """
    Perform convolution of signal with linear time-variant system ``ltv``.

    :param signal: Vector representing input signal :math:`u`.
    :param ltv: 2D array where each column represents an impulse response
    :param mode: 'full', 'valid', or 'same'. See :func:`np.convolve` for an explanation of the options.

    The convolution of two sequences is given by

    .. math:: \mathbf{y} = \mathbf{t} \\star \mathbf{u}

    This can be written as a matrix-vector multiplication

    .. math:: \mathbf{y} = \mathbf{T} \\cdot \mathbf{u}

    where :math:`T` is a Toeplitz matrix in which each column represents an impulse response.
    In the case of a linear time-invariant (LTI) system, each column represents a time-shifted copy of the first column.
    In the time-variant case (LTV), every column can contain a unique impulse response, both in values as in size.

    This function assumes all impulse responses are of the same size.
    The input matrix ``ltv`` thus represents the non-shifted version of the Toeplitz matrix.

    .. seealso:: :func:`np.convolve`, :func:`scipy.signal.convolve` and :func:`scipy.signal.fftconvolve` for convolution with LTI system.

    """
    
    assert(len(signal) == ltv.shape[1])
    
    n = ltv.shape[0] + len(signal) - 1                          # Length of output vector
    un = np.concatenate((signal, np.zeros(ltv.shape[0] - 1)))   # Resize input vector
    offsets = np.arange(0, -ltv.shape[0], -1)                   # Offsets for impulse responses
    Cs = spdiags(ltv, offsets, n, n)                            # Sparse representation of IR's.
    out = Cs.dot(un)                                            # Calculate dot product.

    if mode=='full':
        return out
    elif mode=='same':
        start = ltv.shape[0]/2 - 1 + ltv.shape[0]%2
        stop = len(signal) + ltv.shape[0]/2 - 1 + ltv.shape[0]%2
        return out[start:stop]
    elif mode=='valid':
        length = len(signal) - ltv.shape[0]
        start = ltv.shape[0] - 1
        stop = len(signal)
        return out[start:stop]


def ir2fr(ir, fs, N=None):
    """
    Convert impulse response into frequency response. Returns single-sided RMS spectrum.

    :param ir: Impulser response
    :param fs: Sample frequency
    :param N: Blocks

    Calculates the positive frequencies using :func:`np.fft.rfft`.
    Corrections are then applied to obtain the single-sided spectrum.

    .. note:: Single-sided spectrum. Therefore, the amount of bins returned is either N/2 or N/2+1.

    """
    #ir = ir - np.mean(ir) # Remove DC component.

    N = N if N else ir.shape[-1]
    fr = rfft(ir, n=N) / N
    f = np.fft.rfftfreq(N, 1.0/fs)    #/ 2.0

    fr *= 2.0
    fr[..., 0] /= 2.0    # DC component should not be doubled.
    if not N%2: # if not uneven
        fr[..., -1] /= 2.0 # And neither should fs/2 be.

    #f = np.arange(0, N/2+1)*(fs/N)

    return f, fr


def decibel_to_neper(decibel):
    """
    Convert decibel to neper.

    :param decibel: Value in decibel (dB).
    :returns: Value in neper (Np).

    The conversion is done according to

    .. math :: \\mathrm{dB} = \\frac{\\log{10}}{20} \\mathrm{Np}

    """
    return np.log(10.0) / 20.0  * decibel


def neper_to_decibel(neper):
    """
    Convert neper to decibel.

    :param neper: Value in neper (Np).
    :returns: Value in decibel (dB).

    The conversion is done according to

    .. math :: \\mathrm{Np} = \\frac{20}{\\log{10}} \\mathrm{dB}
    """
    return 20.0 / np.log(10.0) * neper


class Frequencies(object):
    """
    Object describing frequency bands.
    """

    def __init__(self, center, lower, upper, bandwidth=None):

        self.center = np.asarray(center)
        """
        Center frequencies.
        """

        self.lower = np.asarray(lower)
        """
        Lower frequencies.
        """

        self.upper = np.asarray(upper)
        """
        Upper frequencies.
        """

        self.bandwidth = np.asarray(bandwidth) if bandwidth is not None else np.asarray(self.upper) - np.asarray(self.lower)
        """
        Bandwidth.
        """

    def __iter__(self):
        for i in range(len(self.center)):
            yield self[i]

    def __len__(self):
        return len(self.center)

    def __str__(self):
        return str(self.center)

    def __repr__(self):
        return "Frequencies({})".format(str(self.center))

    def angular(self):
        """Angular center frequency in radians per second.
        """
        return 2.0 * np.pi * self.center


class EqualBand(Frequencies):
    """
    Equal bandwidth spectrum. Generally used for narrowband data.
    """

    def __init__(self, center=None, fstart=None, fstop=None, nbands=None, bandwidth=None):
        """

        :param center: Vector of center frequencies.
        :param fstart: First center frequency.
        :param fstop: Last center frequency.
        :param nbands: Amount of frequency bands.
        :param bandwidth: Bandwidth of bands.

        """

        if center is not None:
            try:
                nbands = len(center)
            except TypeError:
                center = [center]
                nbands = 1

            u = np.unique(np.diff(center).round(decimals=3))
            n = len(u)
            if n == 1:
                bandwidth = u
            elif n > 1:
                raise ValueError("Given center frequencies are not equally spaced.")
            else:
                pass
            fstart = center[0] #- bandwidth/2.0
            fstop = center[-1] #+ bandwidth/2.0
        elif fstart is not None and fstop is not None and nbands:
            bandwidth = (fstop - fstart) / (nbands-1)
        elif fstart is not None and fstop is not None and bandwidth:
            nbands = round((fstop - fstart) / bandwidth) + 1
        elif fstart is not None and bandwidth and nbands:
            fstop = fstart + nbands * bandwidth
        elif fstop is not None and bandwidth and nbands:
            fstart = fstop - (nbands-1) * bandwidth
        else:
            raise ValueError("Insufficient parameters. Cannot determine fstart, fstop, bandwidth.")

        center = fstart + np.arange(0, nbands) * bandwidth # + bandwidth/2.0
        upper  = fstart + np.arange(0, nbands) * bandwidth + bandwidth/2.0
        lower  = fstart + np.arange(0, nbands) * bandwidth - bandwidth/2.0

        super(EqualBand, self).__init__(center, lower, upper, bandwidth)

    def __getitem__(self, key):
        return type(self)(center=self.center[key], bandwidth=self.bandwidth)

    def __repr__(self):
        return "EqualBand({})".format(str(self.center))



class OctaveBand(Frequencies):
    """Fractional-octave band spectrum.
    """

    def __init__(self, center=None, fstart=None, fstop=None, nbands=None, fraction=1, reference=acoustics.octave.REFERENCE):


        if center is not None:
            try:
                nbands = len(center)
            except TypeError:
                center = [center]
            center = np.asarray(center)
            indices = acoustics.octave.index_of_frequency(center, fraction=fraction, ref=reference)
        elif fstart is not None and fstop is not None:
            nstart = acoustics.octave.index_of_frequency(fstart, fraction=fraction, ref=reference)
            nstop = acoustics.octave.index_of_frequency(fstop, fraction=fraction, ref=reference)
            indices = np.arange(nstart, nstop+1)
        elif fstart is not None and nbands is not None:
            nstart = acoustics.octave.index_of_frequency(fstart, fraction=fraction, ref=reference)
            indices = np.arange(nstart, nstart+nbands)
        elif fstop is not None and nbands is not None:
            nstop = acoustics.octave.index_of_frequency(fstop, fraction=fraction, ref=reference)
            indices = np.arange(nstop-nbands, nstop)
        else:
            raise ValueError("Insufficient parameters. Cannot determine fstart and/or fstop.")

        center = acoustics.octave.exact_center_frequency(None, fraction=fraction, n=indices, ref=reference)
        lower = acoustics.octave.lower_frequency(center, fraction=fraction)
        upper = acoustics.octave.upper_frequency(center, fraction=fraction)
        bandwidth = upper - lower
        nominal = acoustics.octave.nominal_center_frequency(None, fraction, indices)

        super(OctaveBand, self).__init__(center, lower, upper, bandwidth)

        self.fraction = fraction
        """Fraction of fractional-octave filter.
        """

        self.reference = reference
        """Reference center frequency.
        """

        self.nominal = nominal
        """Nominal center frequencies.
        """

    def __getitem__(self, key):
        return type(self)(center=self.center[key], fraction=self.fraction, reference=self.reference)

    def __repr__(self):
        return "OctaveBand({})".format(str(self.center))


def ms(x):
    """Mean value of signal `x` squared.

    :param x: Dynamic quantity.
    :returns: Mean squared of `x`.

    """
    return (np.abs(x)**2.0).mean()

def rms(x):
    """Root mean squared of signal `x`.

    :param x: Dynamic quantity.

    .. math:: x_{rms} = lim_{T \\to \\infty} \\sqrt{\\frac{1}{T} \int_0^T |f(x)|^2 \\mathrm{d} t }

    :seealso: :func:`ms`.

    """
    return np.sqrt(ms(x))

def normalize(y, x=None):
    """normalize power in y to a (standard normal) white noise signal.

    Optionally normalize to power in signal `x`.

    #The mean power of a Gaussian with :math:`\\mu=0` and :math:`\\sigma=1` is 1.
    """
    #return y * np.sqrt( (np.abs(x)**2.0).mean() / (np.abs(y)**2.0).mean() )
    if x is not None:
        x = ms(x)
    else:
        x = 1.0
    return y * np.sqrt( x / ms(y) )
    #return y * np.sqrt( 1.0 / (np.abs(y)**2.0).mean() )

    ## Broken? Caused correlation in auralizations....weird!




def window_scaling_factor(window, axis=-1):
    """
    Calculate window scaling factor.

    :param window: Window.

    When analysing broadband (filtered noise) signals it is common to normalize
    the windowed signal so that it has the same power as the un-windowed one.

    .. math:: S = \\sqrt{\\frac{\\sum_{i=0}^N w_i^2}{N}}

    """
    return np.sqrt((window*window).mean(axis=axis))


def apply_window(x, window):
    """
    Apply window to signal.

    :param x: Instantaneous signal :math:`x(t)`.
    :param window: Vector representing window.

    :returns: Signal with window applied to it.

    .. math:: x_s(t) = x(t) / S

    where :math:`S` is the window scaling factor.

    .. seealso:: :func:`window_scaling_factor`.

    """
    s = window_scaling_factor(window) # Determine window scaling factor.
    n = len(window)
    windows = x//n  # Amount of windows.
    x = x[0:windows*n] # Truncate final part of signal that does not fit.
    #x = x.reshape(-1, len(window)) # Reshape so we can apply window.
    y = np.tile(window, windows)

    return x * y / s


def amplitude_spectrum(x, fs, N=None):
    """
    Amplitude spectrum of instantaneous signal :math:`x(t)`.

    :param x: Instantaneous signal :math:`x(t)`.
    :param fs: Sample frequency :math:`f_s`.
    :param N: Amount of FFT bins.

    The amplitude spectrum gives the amplitudes of the sinusoidal the signal is built
    up from, and the RMS (root-mean-square) amplitudes can easily be found by dividing
    these amplitudes with :math:`\\sqrt{2}`.

    The amplitude spectrum is double-sided.

    """
    N = N if N else x.shape[-1]
    fr = np.fft.fft(x, n=N) / N
    f = np.fft.fftfreq(N, 1.0/fs)
    return np.fft.fftshift(f), np.fft.fftshift(fr, axes=[-1])


def auto_spectrum(x, fs, N=None):
    """
    Auto-spectrum of instantaneous signal :math:`x(t)`.

    :param x: Instantaneous signal :math:`x(t)`.
    :param fs: Sample frequency :math:`f_s`.
    :param N: Amount of FFT bins.

    The auto-spectrum contains the squared amplitudes of the signal. Squared amplitudes
    are used when presenting data as it is a measure of the power/energy in the signal.

    .. math:: S_{xx} (f_n) = \\overline{X (f_n)} \\cdot X (f_n)

    The auto-spectrum is double-sided.

    """
    f, a = amplitude_spectrum(x, fs, N=N)
    return f, (a*a.conj()).real


def power_spectrum(x, fs, N=None):
    """
    Power spectrum of instantaneous signal :math:`x(t)`.

    :param x: Instantaneous signal :math:`x(t)`.
    :param fs: Sample frequency :math:`f_s`.
    :param N: Amount of FFT bins.

    The power spectrum, or single-sided autospectrum, contains the squared RMS amplitudes of the signal.

    A power spectrum is a spectrum with squared RMS values. The power spectrum is
    calculated from the autospectrum of the signal.

    .. warning:: Does not include scaling to reference value!

    .. seealso:: :func:`auto_spectrum`

    """
    N = N if N else x.shape[-1]
    f, a = auto_spectrum(x, fs, N=N)
    a = a[..., N//2:]
    f = f[..., N//2:]
    a *= 2.0
    a[..., 0] /= 2.0    # DC component should not be doubled.
    if not N%2: # if not uneven
        a[..., -1] /= 2.0 # And neither should fs/2 be.
    return f, a


def angle_spectrum(x, fs, N=None):
    """
    Phase angle spectrum of instantaneous signal :math:`x(t)`.

    :param x: Instantaneous signal :math:`x(t)`.
    :param fs: Sample frequency :math:`f_s`.
    :param N: Amount of FFT bins.

    This function returns a single-sided wrapped phase angle spectrum.

    .. seealso:: :func:`phase_spectrum` for unwrapped phase spectrum.

    """
    N = N if N else x.shape[-1]
    f, a = amplitude_spectrum(x, fs, N)
    a = np.angle(a)
    a = a[..., N//2:]
    f = f[..., N//2:]
    return f, a


def phase_spectrum(x, fs, N=None):
    """
    Phase spectrum of instantaneous signal :math:`x(t)`.

    :param x: Instantaneous signal :math:`x(t)`.
    :param fs: Sample frequency :math:`f_s`.
    :param N: Amount of FFT bins.

    This function returns a single-sided unwrapped phase spectrum.

    .. seealso:: :func:`angle_spectrum` for wrapped phase angle.

    """
    f, a = angle_spectrum(x, fs, N=None)
    return f, np.unwrap(a)



#def power_and_phase_spectrum(x, fs, N=None):
    #"""
    #Power spectrum and phase of instantaneous signal :math:`x(t)`.

    #:param x: Instantaneous signal :math:`x(t)`.
    #:param fs: Sample frequency :math:`f_s`.
    #:param N: Amount of FFT bins.

    #Often one is interested in both the power spectrum and the phase. This function returns the power and a single-sided phase spectrum.
    #For an explanation of the power spectrum, see :func:`power_spectrum`.
    #"""


    #returns f, power, phase



def density_spectrum(x, fs, N=None):
    """
    Density spectrum of instantaneous signal :math:`x(t)`.

    :param x: Instantaneous signal :math:`x(t)`.
    :param fs: Sample frequency :math:`f_s`.
    :param N: Amount of FFT bins.

    A density spectrum considers the amplitudes per unit frequency.
    Density spectra are used to compare spectra with different frequency resolution as the
    magnitudes are not influenced by the resolution because it is per Hertz. The amplitude
    spectra on the other hand depend on the chosen frequency resolution.

    """
    N = N if N else x.shape[-1]
    fr = np.fft.fft(x, n=N) / fs
    f = np.fft.fftfreq(N, 1.0/fs)
    return np.fft.fftshift(f), np.fft.fftshift(fr)

#def auto_density_spectrum(x, fs, N=None):
    #"""
    #Auto density spectrum of instantaneous signal :math:`x(t)`.
    #"""
    #f, d = density_spectrum(x, fs, N=N)
    #return f, (d*d.conj()).real

#def power_density_spectrum(x, fs, N=None):
    #"""
    #Power density spectrum.
    #"""
    #N = N if N else x.shape[-1]
    #f, a = auto_density_spectrum(x, fs, N=N)
    #a = a[N//2:]
    #f = f[N//2:]
    #a *= 2.0
    #a[..., 0] /= 2.0    # DC component should not be doubled.
    #if not N%2: # if not uneven
        #a[..., -1] /= 2.0 # And neither should fs/2 be.
    #return f, a


def integrate_bands(data, a, b):
    """
    Reduce frequency resolution of power spectrum. Merges frequency bands by integration.

    :param data: Vector with narrowband powers.
    :param a: Instance of :class:`Frequencies`.
    :param b: Instance of :class:`Frequencies`.

    .. note:: Needs rewriting so that the summation goes over axis=1.

    """

    try:
        if b.fraction%a.fraction:
            raise NotImplementedError("Non-integer ratio of fractional-octaves are not supported.")
    except AttributeError:
        pass

    lower, _ = np.meshgrid(b.lower, a.center)
    upper, _ = np.meshgrid(b.upper, a.center)
    _, center= np.meshgrid(b.center, a.center)

    return ((lower < center) * (center <= upper) * data[...,None]).sum(axis=-2)


def bandpass_frequencies(x, fs, frequencies, order=8, purge=False, zero_phase=False):
    """"Apply bandpass filters for frequencies

    :param x: Instantaneous signal :math:`x(t)`.
    :param fs: Sample frequency.
    :param frequencies: Frequencies. Instance of :class:`Frequencies`.
    :param order: Filter order.
    :param purge: Discard bands of which the upper corner frequency is above the Nyquist frequency.
    :param zero_phase: Prevent phase error by filtering in both directions (filtfilt)
    :returns: Tuple. First element is an instance of :class:`OctaveBand`. The second element an array.
    """
    if purge:
        frequencies = frequencies[frequencies.upper < fs/2.0]
    return frequencies, np.array([bandpass(x, band.lower, band.upper, fs, order, zero_phase=zero_phase) for band in frequencies])


def bandpass_octaves(x, fs, frequencies=NOMINAL_OCTAVE_CENTER_FREQUENCIES, order=8, purge=False, zero_phase=False):
    """Apply 1/1-octave bandpass filters.

    :param x: Instantaneous signal :math:`x(t)`.
    :param fs: Sample frequency.
    :param frequencies: Frequencies.
    :param order: Filter order.
    :param purge: Discard bands of which the upper corner frequency is above the Nyquist frequency.
    :param zero_phase: Prevent phase error by filtering in both directions (filtfilt)
    :returns: Tuple. First element is an instance of :class:`OctaveBand`. The second element an array.

    .. seealso:: :func:`octavepass`
    """
    return bandpass_fractional_octaves(x, fs, frequencies, fraction=1, order=order, purge=purge, zero_phase=zero_phase)


def bandpass_third_octaves(x, fs, frequencies=NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES, order=8, purge=False, zero_phase=False):
    """Apply 1/3-octave bandpass filters.

    :param x: Instantaneous signal :math:`x(t)`.
    :param fs: Sample frequency.
    :param frequencies: Frequencies.
    :param order: Filter order.
    :param purge: Discard bands of which the upper corner frequency is above the Nyquist frequency.
    :param zero_phase: Prevent phase error by filtering in both directions (filtfilt)
    :returns: Tuple. First element is an instance of :class:`OctaveBand`. The second element an array.

    .. seealso:: :func:`octavepass`
    """
    return bandpass_fractional_octaves(x, fs, frequencies, fraction=3, order=order, purge=purge, zero_phase=zero_phase)


def bandpass_fractional_octaves(x, fs, frequencies, fraction=None, order=8, purge=False, zero_phase=False):
    """Apply 1/N-octave bandpass filters.

    :param x: Instantaneous signal :math:`x(t)`.
    :param fs: Sample frequency.
    :param frequencies: Frequencies. Either instance of :class:`OctaveBand`, or array along with fs.
    :param order: Filter order.
    :param purge: Discard bands of which the upper corner frequency is above the Nyquist frequency.
    :param zero_phase: Prevent phase error by filtering in both directions (filtfilt)
    :returns: Tuple. First element is an instance of :class:`OctaveBand`. The second element an array.

    .. seealso:: :func:`octavepass`
    """
    if not isinstance(frequencies, Frequencies):
        frequencies = OctaveBand(center=frequencies, fraction=fraction)
    return bandpass_frequencies(x, fs, frequencies, order=order, purge=purge, zero_phase=zero_phase)



def third_octaves(p, fs, density=False,
                  frequencies=NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES,
                  ref=REFERENCE_PRESSURE):
    """Calculate level per 1/3-octave in frequency domain using the FFT.

    :param x: Instantaneous signal :math:`x(t)`.
    :param fs: Sample frequency.
    :param density: Power density instead of power.
    :returns: Tuple. First element is an instance of :class:`OctaveBand`. The second element an array.

    .. note:: Based on power spectrum (FFT)

    .. seealso:: :attr:`acoustics.bands.THIRD_OCTAVE_CENTER_FREQUENCIES`

    .. note:: Exact center frequencies are always calculated.

    """
    fob = OctaveBand(center=frequencies, fraction=3)
    f, p = power_spectrum(p, fs)
    fnb = EqualBand(f)
    power = integrate_bands(p, fnb, fob)
    if density:
        power /= (fob.bandwidth/fnb.bandwidth)
    level = 10.0*np.log10(power / ref**2.0)
    return fob, level


def octaves(p, fs, density=False,
            frequencies=NOMINAL_OCTAVE_CENTER_FREQUENCIES,
            ref=REFERENCE_PRESSURE):
    """Calculate level per 1/1-octave in frequency domain using the FFT.

    :param x: Instantaneous signal :math:`x(t)`.
    :param fs: Sample frequency.
    :param density: Power density instead of power.
    :param frequencies: Frequencies.
    :param ref: Reference value.
    :returns: Tuple. First element is an instance of :class:`OctaveBand`. The second element an array.

    .. note:: Based on power spectrum (FFT)

    .. seealso:: :attr:`acoustics.bands.OCTAVE_CENTER_FREQUENCIES`

    .. note:: Exact center frequencies are always calculated.

    """
    fob = OctaveBand(center=frequencies, fraction=1)
    f, p = power_spectrum(p, fs)
    fnb = EqualBand(f)
    power = integrate_bands(p, fnb, fob)
    if density:
        power /= (fob.bandwidth/fnb.bandwidth)
    level = 10.0*np.log10(power / ref**2.0)
    return fob, level


def fractional_octaves(p, fs, start=5.0, stop=16000.0, fraction=3, density=False):
    """Calculate level per 1/N-octave in frequency domain using the FFT. N is `fraction`.

    :param x: Instantaneous signal :math:`x(t)`.
    :param fs: Sample frequency.
    :param density: Power density instead of power.
    :returns: Tuple. First element is an instance of :class:`OctaveBand`. The second element an array.

    .. note:: Based on power spectrum (FFT)

    .. note:: This function does *not* use nominal center frequencies.

    .. note:: Exact center frequencies are always calculated.
    """
    fob = OctaveBand(fstart=start, fstop=stop, fraction=fraction)
    f, p = power_spectrum(p, fs)
    fnb = EqualBand(f)
    power = integrate_bands(p, fnb, fob)
    if density:
        power /= (fob.bandwidth/fnb.bandwidth)
    level = 10.0*np.log10(power)
    return fob, level


class Filterbank(object):
    """
    Fractional-Octave filter bank.


    .. warning:: For high frequencies the filter coefficients are wrong for low frequencies. Therefore, to improve the response for lower frequencies the signal should be downsampled. Currently, there is no easy way to do so within the Filterbank.

    """

    def __init__(self, frequencies, sample_frequency=44100, order=8):


        self.frequencies = frequencies
        """
        Frequencies object.

        See also :class:`Frequencies` and subclasses.

        .. note:: A frequencies object should have the attributes center, lower and upper.

        """

        self.order = order
        """
        Filter order of Butterworth filter.
        """

        self.sample_frequency = sample_frequency
        """
        Sample frequency.
        """

    @property
    def sample_frequency(self):
        """
        Sample frequency.
        """
        return self._sample_frequency

    @sample_frequency.setter
    def sample_frequency(self, x):
        #if x <= self.center_frequencies.max():
            #raise ValueError("Sample frequency cannot be lower than the highest center frequency.")
        self._sample_frequency = x

    @property
    def filters(self):
        """
        Filters this filterbank consists of.
        """
        fs = self.sample_frequency
        return ( bandpass_filter(lower, upper, fs, order=self.order, output='sos') for lower, upper in zip(self.frequencies.lower, self.frequencies.upper) )

        #order = self.order
        #filters = list()
        #nyq = self.sample_frequency / 2.0
        #return ( butter(order, [lower/nyq, upper/nyq], btype='band', analog=False) for lower, upper in zip(self.frequencies.lower, self.frequencies.upper) )

    def lfilter(self, signal):
        """
        Filter signal with filterbank.

        .. note:: This function uses :func:`scipy.signal.lfilter`.
        """
        return ( sosfilt(sos, signal) for sos in self.filters )

    def filtfilt(self, signal):
        """
        Filter signal with filterbank.
        Returns a list consisting of a filtered signal per filter.

        .. note:: This function uses :func:`scipy.signal.filtfilt` and therefore has a zero-phase response.
        """
        return ( _sosfiltfilt(sos, signal) for sos in self.filters )

    def power(self, signal):
        """
        Power per band in signal.
        """
        filtered = self.filtfilt(signal)
        return np.array([(x**2.0).sum()/len(x) / bw for x, bw in zip(filtered, self.frequencies.bandwidth)])

    def plot_response(self):
        """
        Plot frequency response.

        .. note:: The follow phase response is obtained in case :meth:`lfilter` is used. The method :meth:`filtfilt` results in a zero-phase response.
        """

        fs = self.sample_frequency
        fig = plt.figure()
        ax1 = fig.add_subplot(211)
        ax2 = fig.add_subplot(212)
        for f, fc in zip(self.filters, self.frequencies.center):
            w, h = freqz(f[0], f[1], int(fs/2))#np.arange(fs/2.0))
            ax1.semilogx(w / (2.0*np.pi) * fs, 20.0 * np.log10(np.abs(h)), label=str(int(fc)))
            ax2.semilogx(w / (2.0*np.pi) * fs, np.angle(h), label=str(int(fc)))
        ax1.set_xlabel(r'$f$ in Hz')
        ax1.set_ylabel(r'$|H|$ in dB re. 1')
        ax2.set_xlabel(r'$f$ in Hz')
        ax2.set_ylabel(r'$\angle H$ in rad')
        ax1.legend(loc=5)
        ax2.legend(loc=5)
        ax1.set_ylim(-60.0, +10.0)

        return fig

    def plot_power(self, signal):
        """
        Plot power in signal.
        """

        f = self.frequencies.center
        p = self.power(signal)

        fig = plt.figure()
        ax = fig.add_subplot(111)
        p = ax.bar(f, 20.0*np.log10(p))
        ax.set_xlabel('$f$ in Hz')
        ax.set_ylabel('$L$ in dB re. 1')
        ax.set_xscale('log')

        return fig


#class FilterbankFFT(object):
    #"""
    #Filterbank to filter signal using FFT.
    #"""

    #def __init__(self, frequencies, sample_frequency=44100):

       #self.frequencies = frequencies
       #"""
       #Frequencies.

       #See also :class:`Frequencies` and subclasses.
       #"""
       #self.sample_frequency = sample_frequency


    #def power(self, signal):
        #pass

    #def plot_power(self, signal):
        #pass



def isolate(signals):
    """Isolate signals.

    :param signals: Array of shape N x M where N is the amount of samples and M the amount of signals. Thus, each column is a signal.
    :returns: Array of isolated signals. Each column is a signal.

    Isolate signals using Singular Value Decomposition.

    """
    x = np.asarray(signals)

    W, s, v = np.linalg.svd( (np.tile( (x*x).sum(axis=0), (len(x), 1) ) * x).dot(x.T) )
    return v.T


def zero_crossings(data):
    """
    Determine the positions of zero crossings in `data`.

    :param data: Vector

    :returns: Vector with indices of samples *before* the zero crossing.

    """
    pos = data > 0
    npos = ~pos
    return ((pos[:-1] & npos[1:]) | (npos[:-1] & pos[1:])).nonzero()[0]


def amplitude_envelope(signal, fs, axis=-1):
    """Instantaneous amplitude of tone.

    The instantaneous amplitude is the magnitude of the analytic signal.

    :param signal: Signal.
    :param fs: Sample frequency.
    :param axis: Axis.
    :returns: Amplitude envelope of `signal`.

    .. seealso:: :func:`scipy.signal.hilbert`

    """
    return np.abs(hilbert(signal, axis=axis))


def instantaneous_phase(signal, fs, axis=-1):
    """Instantaneous phase of tone.

    :param signal: Signal.
    :param fs: Sample frequency.
    :param axis: Axis.
    :returns: Instantaneous phase of `signal`.

    The instantaneous phase is the angle of the analytic signal.
    This function returns a wrapped angle.

    .. seealso:: :func:`scipy.signal.hilbert`

    """
    return np.angle(hilbert(signal, axis=axis))


def instantaneous_frequency(signal, fs, axis=-1):
    """Determine instantaneous frequency of tone.

    :param signal: Signal.
    :param fs: Sample frequency.
    :param axis: Axis.
    :returns: Instantaneous frequency of `signal`.

    The instantaneous frequency can be obtained by differentiating the unwrapped instantaneous phase.

    .. seealso:: :func:`instantaneous_phase`

    """
    return np.diff( np.unwrap(instantaneous_phase(signal, fs, axis=axis), axis=axis), axis=axis) / (2.0*np.pi) * fs


def wvd(signal, fs, analytic=True):
    """Wigner-Ville Distribution

    :param signal: Signal
    :param fs: Sample frequency
    :param analytic: Use the analytic signal, calculated using Hilbert transform.

    .. math:: W_z(n, \\omega) = 2 \\sum_k z^*[n-k]z[n+k] e^{-j\\omega 2kT}

    Includes positive and negative frequencies.

    """
    signal = np.asarray(signal)

    N = int(len(signal)+len(signal)%2)
    length_FFT = N # Take an even value of N

    #if N != len(signal):
    #    signal = np.concatenate(signal, [0])

    length_time = len(signal)

    if analytic:
        signal = hilbert(signal)
    s = np.concatenate((np.zeros(length_time), signal, np.zeros(length_time)))
    W = np.zeros((length_FFT,length_time))
    tau = np.arange(0, N//2)

    R = np.zeros((N, length_time), dtype='float64')

    i = length_time
    for t in range(length_time):
        R[t, tau1] = ( s[i+tau] * s[i-tau].conj() ) # In one direction
        R[t, N-(tau+1)] = R[t, tau+1].conj() # And the other direction
        i += 1
    W = np.fft.fft(R, length_FFT) / (2*length_FFT)

    f = np.fft.fftfreq(N, 1./fs)
    return f, W.T


def _sosfiltfilt(sos, x, axis=-1, padtype='odd', padlen=None, method='pad', irlen=None):
    """Filtfilt version using Second Order sections. Code is taken from scipy.signal.filtfilt and adapted to make it work with SOS.
    Note that broadcasting does not work.
    """
    from scipy.signal import sosfilt_zi
    from scipy.signal._arraytools import odd_ext, axis_slice, axis_reverse
    x = np.asarray(x)

    if padlen is None:
        edge = 0
    else:
        edge = padlen

    # x's 'axis' dimension must be bigger than edge.
    if x.shape[axis] <= edge:
        raise ValueError("The length of the input vector x must be at least "
                         "padlen, which is %d." % edge)

    if padtype is not None and edge > 0:
        # Make an extension of length `edge` at each
        # end of the input array.
        if padtype == 'even':
            ext = even_ext(x, edge, axis=axis)
        elif padtype == 'odd':
            ext = odd_ext(x, edge, axis=axis)
        else:
            ext = const_ext(x, edge, axis=axis)
    else:
        ext = x

    # Get the steady state of the filter's step response.
    zi = sosfilt_zi(sos)

    # Reshape zi and create x0 so that zi*x0 broadcasts
    # to the correct value for the 'zi' keyword argument
    # to lfilter.
    #zi_shape = [1] * x.ndim
    #zi_shape[axis] = zi.size
    #zi = np.reshape(zi, zi_shape)
    x0 = axis_slice(ext, stop=1, axis=axis)
    # Forward filter.
    (y, zf) = sosfilt(sos, ext, axis=axis, zi=zi * x0)

    # Backward filter.
    # Create y0 so zi*y0 broadcasts appropriately.
    y0 = axis_slice(y, start=-1, axis=axis)
    (y, zf) = sosfilt(sos, axis_reverse(y, axis=axis), axis=axis, zi=zi * y0)

    # Reverse y.
    y = axis_reverse(y, axis=axis)

    if edge > 0:
        # Slice the actual signal from the extended signal.
        y = axis_slice(y, start=edge, stop=-edge, axis=axis)

    return y

from scipy.signal import lti, cheby1, firwin

def decimate(x, q, n=None, ftype='iir', axis=-1, zero_phase=False):
    """
    Downsample the signal by using a filter.

    By default, an order 8 Chebyshev type I filter is used.  A 30 point FIR
    filter with hamming window is used if `ftype` is 'fir'.

    Parameters
    ----------
    x : ndarray
        The signal to be downsampled, as an N-dimensional array.
    q : int
        The downsampling factor.
    n : int, optional
        The order of the filter (1 less than the length for 'fir').
    ftype : str {'iir', 'fir'}, optional
        The type of the lowpass filter.
    axis : int, optional
        The axis along which to decimate.
    zero_phase : bool
        Prevent phase shift by filtering with ``filtfilt`` instead of ``lfilter``.
    Returns
    -------
    y : ndarray
        The down-sampled signal.

    See also
    --------
    resample

    Notes
    -----
    The ``zero_phase`` keyword was added in 0.17.0.
    The possibility to use instances of ``lti`` as ``ftype`` was added in 0.17.0.

    """

    if not isinstance(q, int):
        raise TypeError("q must be an integer")

    if ftype == 'fir':
        if n is None:
            n = 30
        system = lti(firwin(n + 1, 1. / q, window='hamming'), 1.)

    elif ftype == 'iir':
        if n is None:
            n = 8
        system = lti(*cheby1(n, 0.05, 0.8 / q))
    else:
        system = ftype

    if zero_phase:
        y = filtfilt(system.num, system.den, x, axis=axis)
    else:
        y = lfilter(system.num, system.den, x, axis=axis)

    sl = [slice(None)] * y.ndim
    sl[axis] = slice(None, None, q)
    return y[sl]


def impulse_response_real_even(tf, ntaps):
    """The impulse response of a real and even frequency response is also real and even.

    :param tf: Real and even frequency response. Only positive frequencies.
    :param ntaps: Amount of taps.
    :returns: A real and even (double-sided) impulse response with length `ntaps`.

    A symmetric impulse response is needed. The center of symmetry determines the delay
    of the filter and thereby whether the filter is causal (delay>0, linear-phase) or
    non-causal (delay=0, linear-phase, zero-phase).

    Creating linear phase can be done by multiplying the magnitude with a complex
    exponential corresponding to the desired shift. Another method is to rotate the
    impulse response.

    https://ccrma.stanford.edu/~jos/filters/Zero_Phase_Filters_Even_Impulse.html
    """
    ir = np.fft.ifftshift(np.fft.irfft(tf, n=ntaps)).real
    return ir


def linear_phase(ntaps, steepness=1):
    """Compute linear phase delay for a single-sided spectrum.

    :param ntaps: Amount of filter taps.
    :param steepness: Steepness of phase delay. Default value is 1, corresponding to delay in samples of `ntaps//2`.
    :returns: Linear phase delay.

    A linear phase delay can be added to an impulse response using the function `np.fft.ifftshift`.
    Sometimes, however, you would like to add the linear phase delay to the frequency response instead.
    This function computes the linear phase delay which can be multiplied with a single-sided frequency response.
    """
    f = np.fft.rfftfreq(ntaps, 1.0) # Frequencies normalized to Nyquist.
    alpha = ntaps//2 * steepness
    return np.exp(-1j*2.*np.pi*f*alpha)


__all__ = ['bandpass',
           'bandpass_frequencies',
           'bandpass_fractional_octaves',
           'bandpass_octaves',
           'bandpass_third_octaves',
           'lowpass',
           'highpass',
           'octavepass',
           'octave_filter',
           'bandpass_filter',
           'convolve',
           'ir2fr',
           'decibel_to_neper',
           'neper_to_decibel',
           'EqualBand',
           'OctaveBand',
           'ms',
           'rms',
           'normalize',
           'window_scaling_factor',
           'apply_window',
           'amplitude_spectrum',
           'auto_spectrum',
           'power_spectrum',
           'angle_spectrum',
           'phase_spectrum',
           'density_spectrum',
           'integrate_bands',
           'octaves',
           'third_octaves',
           'fractional_octaves',
           'Filterbank',
           'isolate',
           'zero_crossings',
           'amplitude_envelope',
           'instantaneous_phase',
           'instantaneous_frequency',
           'wvd',
           'decimate',
           ]



"""
Reflection
==========

The reflection module contains functions for calculating reflection factors and impedances.
"""
from __future__ import division

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import AxesGrid, make_axes_locatable

from scipy.special import erfc


SPECIFIC_HEAT_RATIO = 1.4
"""Specific heat ratio of air :math:`\\gamma`."""
POROSITY_DECREASE = 120.0
"""Rate of exponential decrease of porosity with depth :math:`\\alpha`."""
SOUNDSPEED = 343.0
"""Speed of sound in air :math:`c`."""
DENSITY = 1.296
"""Density of air :math:`\\rho`."""

class Boundary(object):
    """
    An object describing a boundary.
    """
    
    def __init__(self, frequency, flow_resistivity, density=DENSITY, soundspeed=SOUNDSPEED, porosity_decrease=POROSITY_DECREASE, specific_heat_ratio=SPECIFIC_HEAT_RATIO, angle=None, distance=None, impedance_model='db', reflection_model='plane'):
        
        
        self.frequency = frequency
        """
        Frequency. Single value or vector for a frequency range.
        """
        
        self.flow_resistivity = flow_resistivity
        """
        Flow resistivity :math:`\\sigma`.
        """
        
        self.density = density
        """
        Density of air :math:`\\rho`.
        
        .. note:: This value is only required for when calculating the impedance according to Attenborough's model. See :func:`impedance_attenborough`.
        
        """
        
        self.soundspeed = soundspeed
        """
        Speed of sound in air :math:`c`.
        
        .. note:: This value is required when calculating the impedance according to Attenborough's model or when calculating the spherical wave reflection factor. See respectively :func:`impedance_attenborough` and :func:`reflection_factor_spherical_wave`.
        
        """
        
        self.porosity_decrease = porosity_decrease
        """
        Rate of exponential decrease of porosity with depth :math:`\\alpha`.
        
        .. note:: This value is only required for when calculating the impedance according to Attenborough's model. See :func:`impedance_attenborough`.
        
        """
        
        self.specific_heat_ratio = specific_heat_ratio
        """
        Ratio of specific heats :math:`\\gamma` for air.
        
        .. note:: This value is only required for when calculating the impedance according to Attenborough's model. See :func:`impedance_attenborough`.
        
        """
        
        self.angle = angle
        """
        Angle of incidence :math:`\\theta`.
        
        """
        
        self.distance = distance
        """
        Path length of the reflected ray :math:`r`.
        
        .. note:: This value is only required for when calculating the spherical wave reflection factor. See :func:`reflection_factor_spherical_wave`.
        
        """

        self.impedance_model = impedance_model
        """
        Impedance model.
        
        Possibilities are ``'db'`` and ``'att'`` for respectively :func:`impedance_delany_and_bazley` and :func:`impedance_attenborough`.
        """
        
        self.reflection_model = reflection_model
        """
        Reflection factor model.
        
        Possibilities are ``'plane'`` and ``'spherical'``` for respectively :func:`reflection_factor_plane_wave` and :func:`reflection_factor_spherical_wave`.
        """

    @property
    def wavenumber(self):
        """
        Wavenumber :math:`k`.
        
        .. math:: k = \\frac{2 \\pi f}{c}
        """
        return 2.0 * np.pi * self.frequency / self.soundspeed

    @property
    def impedance(self):
        """
        Impedance according to chosen impedance model defined using :meth:`impedance_model`.
        """
        if self.impedance_model=='db':
            return impedance_delany_and_bazley(self.frequency, self.flow_resistivity)
        if self.impedance_model=='att':
            return impedance_attenborough(self.frequency, self.flow_resistivity, self.density, self.soundspeed, self.porosity_decrease, self.specific_heat_ratio)

    @property
    def reflection_factor(self):
        """
        Reflection factor according to chosen reflection factor model defined using :meth:`reflection_model`.
        """
        if self.angle is None:
            raise AttributeError('Cannot calculate reflection factor. self.angle has not been specified.')
        
        
        
        if self.reflection_model=='plane':
            return reflection_factor_plane_wave(*np.meshgrid(self.impedance, self.angle))
        elif self.reflection_model=='spherical':
            if self.distance is None:
                raise AttributeError('Cannot calculate reflection factor. self.distance has not been specified.')
            else:
                return reflection_factor_spherical_wave(*np.meshgrid(self.impedance, self.angle), distance=self.distance, wavenumber=self.wavenumber)
        else:
            raise RuntimeError("Oops...")
        
        
    def plot_impedance(self, filename=None):
        """
        Plot magnitude and phase of the impedance as function of frequency.
        """
        
        fig = plt.figure()
        
        ax0 = fig.add_subplot(211)
        ax0.set_title('Magnitude of impedance')
        ax0.semilogx(self.frequency, np.abs(self.impedance))
        ax0.set_xlabel(r'$f$ in Hz')
        ax0.set_ylabel(r'$\left|Z\right|$') 
        ax0.grid()
        
        ax0 = fig.add_subplot(212)
        ax0.set_title('Angle of impedance')
        ax0.semilogx(self.frequency, np.angle(self.impedance))
        ax0.set_xlabel(r'$f$ in Hz')
        ax0.set_ylabel(r'$\angle Z$') 
        ax0.grid()
        
        plt.tight_layout()
        
        if filename:
            fig.savefig(filename, transparant=True)
        else:
            return fig
    
    
    def plot_reflection_factor(self, filename=None):
        """
        Plot reflection factor.
        """
        
        if self.frequency is None:
            raise ValueError("No frequency specified.")
        if self.angle is None:
            raise ValueError("No angle specified.")
        
        try:
            n_f = len(self.frequency)
        except TypeError:
            n_f = 1
        try:
            n_a = len(self.angle)
        except TypeError:
            n_a = 1
        
        
        if n_f==1 and n_a==1:
            raise ValueError("Either frequency or angle needs to be a vector.")
        
        elif n_f==1 or n_a==1:
            if n_f==1 and n_a>1:# Show R as function of angle for a single frequency.
                xlabel = r"$\theta$ in degrees"
            elif n_f>1 and n_a==1:# Show R as function of frequency for a single angle.
                xlabel = r"$f$ in Hz"
            R = self.reflection_factor
            fig = plt.figure()
                
            ax0 = fig.add_subplot(211)
            ax0.set_title("Magnitude of reflection factor")
            ax0.semilogx(self.frequency, np.abs(R))
            ax0.set_xlabel(xlabel)
            ax0.set_ylabel(r'$\left|R\right|$') 
            ax0.grid()
            
            ax1 = fig.add_subplot(212)
            ax1.set_title("Phase of reflection factor")
            ax1.semilogx(self.frequency, np.angle(R))
            ax1.set_xlabel(xlabel)
            ax1.set_ylabel(r'$\angle R$') 
            ax1.grid()

            
        elif n_f>1 and n_a>1:# Show 3D or pcolor
            R = self.reflection_factor
            fig = plt.figure()
            
            #grid = AxesGrid(fig, 111, nrows_ncols=(2, 2), axes_pad=0.1, cbar_mode='each', cbar_location='right')
            ax0 = fig.add_subplot(211)
            #ax0 = grid[0]
            ax0.set_title("Magnitude of reflection factor")
            ax0.pcolormesh(self.frequency, self.angle * 180.0 / np.pi, np.abs(R))
            #ax0.pcolor(self.angle, self.frequency, np.abs(R))
            #ax0.set_xlabel(xlabel)
            #ax0.set_ylabel(r'$\left|R\right|$') 
            ax0.grid()
            
            ax1 = fig.add_subplot(212)
            #ax1 = grid[1]
            ax1.set_title("Phase of reflection factor")
            ax1.pcolormesh(self.frequency, self.angle * 180.0 / np.pi, np.angle(R))
            #ax1.pcolor(self.angle, self.frequency, np.angle(R))
            #ax0.set_xlabel(xlabel)
            #ax0.set_ylabel(r'$\angle R$') 
            ax1.grid()

        else:
            raise RuntimeError("Oops...")
        
        #plt.tight_layout()
        
        if filename:
            fig.savefig(filename, transparant=True)
        else:
            return fig
    
    
  
def reflection_factor_plane_wave(impedance, angle):
    """
    Plane wave reflection factor :math:`R`.
    
    :param impedance: Normalized impedance :math:`Z`.
    :param angle: Angle of incidence :math:`\\theta`.
    
    
    The plane wave reflection factor :math:`R` is given by
    
    .. math:: R = \\frac{Z\cos{\\theta}-1}{Z\cos{\\theta}+1}
    
    where :math:`Z` is the normalized impedance and :math:`\\theta` the angle of incidence.
    
    """
    return (impedance * np.cos(angle) - 1.0) / (impedance * np.cos(angle) + 1.0)

def numerical_distance(impedance, angle, distance, wavenumber):
    """
    Numerical distance :math:`w`.
    
    :param impedance: Normalized impedance :math:`Z`.
    :param angle: Angle of incidence :math:`\\theta`.
    :param distance: Path length of the reflected ray :math:`r`.
    :param wavenumber: Wavenumber :math:`k`.    
    
    The numerical distance :math:`w` is given by
    
    .. math:: w = \\sqrt{-j k r  \\left( 1 + \\frac{1}{Z} \cos{\\theta} - \\sqrt{1 - \\left( \\frac{1}{Z} \\right)^2} \\sin{\\theta} \\right) }
    
    """
    return np.sqrt(-1j*wavenumber*distance*(1.0 + 1.0/impedance * np.cos(angle) - np.sqrt(1.0-(1.0/impedance)**2.0)*np.sin(angle)))


def reflection_factor_spherical_wave(impedance, angle, distance, wavenumber):
    """
    Spherical wave reflection factor :math:`Q`.
    
    :param impedance: Normalized impedance :math:`Z`.
    :param angle: Angle of incidence :math:`\\theta`.
    :param distance: Path length of the reflected ray :math:`r`.
    :param wavenumber: Wavenumber :math:`k`.  
    
    The spherical wave relfection factor :math:`Q` is given by
    
    .. math:: Q = R \\left(1 - R \\right) F
    
    where :math:`R` is the plane wave reflection factor as calculated in :func:`reflection_factor_plane_wave` and :math:`F` is given by
    
    .. math:: F = 1 - j \\sqrt{ \\pi} w e^{-w^2} \mathrm{erfc} \\left( j w \\right) 
    
    where :math:`w` is the numerical distance as calculated in :func:`numerical_distance`.
    """
    w = numerical_distance(impedance, angle, distance, wavenumber)

    F = 1.0 - 1j * np.sqrt(np.pi)*w*np.exp(-w**2.0) * erfc(1j*w)

    plane_factor = reflection_factor_plane_wave(impedance, angle)
    return plane_factor * (1.0 - plane_factor) * F


def impedance_delany_and_bazley(frequency, flow_resistivity):
    """
    Normalised specific acoustic impedance according to the empirical one-parameter model by Delany and Bazley.
    
    :param frequency: Frequency :math:`f`.
    :param flow_resistivity: Flow resistivity :math:`\\sigma`.
    
    The impedance :math:`Z` is given by
    
    .. math:: Z = 1 + 9.08 \\left( \\frac{1000f}{\\sigma}\\right)^{-0.75} - 11.9 j \\left( \\frac{1000f}{\\sigma}\\right)^{-0.73}
    
    """
    return 1.0 + 9.08 * (1000.0*frequency/flow_resistivity)**(-0.75) - 1j * 11.9 * (1000.0*frequency/flow_resistivity)**(-0.73)
    
    
def impedance_attenborough(frequency, flow_resistivity, density=DENSITY, soundspeed=SOUNDSPEED, porosity_decrease=POROSITY_DECREASE, specific_heat_ratio=SPECIFIC_HEAT_RATIO):
    """
    Normalised specific acoustics impedance according to the two-parameter model by Attenborough.
    
    :param frequency: Frequency :math:`f`.
    :param flow_resistivity: Flow resistivity :math:`\\sigma`.
    :param soundspeed: Speed of sound in air :math:`c`.
    :param density: Density of air :math:`\\rho`.
    :param porosity_decrease: Rate of exponential decrease of porosity with depth :math:`\\alpha`.
    :param specific_heat_ratio: Ratio of specific heats :math:`\\gamma` for air.
    
    
    The impedance :math:`Z` is given by
    
    .. math:: Z = \\frac{\\left( 1-j\\right) \\sqrt{\\sigma/f}}{\\sqrt{\\pi \\gamma_0 \\rho_0}} - \\frac{jc\\alpha}{8 \\pi \\gamma_0 f}
    
    """
    return (1.0 - 1j) * np.sqrt(flow_resistivity/frequency)/np.sqrt(np.pi*specific_heat_ratio*density) - 1j*soundspeed*porosity_decrease/(8.0*np.pi*specific_heat_ratio*frequency)



"""
ISO/TR 25417 2007
=================

ISO/TR 25417:2007 specifies definitions of acoustical quantities and terms used
in noise measurement documents prepared by ISO Technical Committee TC 43,
Acoustics, Subcommittee SC 1, Noise, together with their symbols and units, with
the principal aim of harmonizing the terminology used [ISO25417]_.

.. [ISO24517] http://www.iso.org/iso/home/store/catalogue_tc/catalogue_detail.htm?csnumber=42915

.. inheritance-diagram:: acoustics.standards.iso_tr_25417_2007

"""
import numpy as np

REFERENCE_PRESSURE = 2.0e-5
"""
Reference value of the sound pressure :math:`p_0` is :math:`2 \cdot 10^{-5}` Pa.
"""

def sound_pressure_level(pressure, reference_pressure=REFERENCE_PRESSURE):
    """
    Sound pressure level :math:`L_p` in dB.

    :param pressure: Instantaneous sound pressure :math:`p`.
    :param reference_pressure: Reference value :math:`p_0`.

    .. math:: L_p = 10 \\log_{10}{ \\left( \\frac{p^2}{p_0^2} \\right)}

    See section 2.2.
    """
    return 10.0 * np.log10( pressure**2.0 / reference_pressure**2.0 )


def equivalent_sound_pressure_level(pressure, reference_pressure=REFERENCE_PRESSURE, axis=-1):
    """
    Time-averaged sound pressure level :math:`L_{p,T}` or equivalent-continious sound pressure level :math:`L_{p,eqT}` in dB.

    :param pressure: Instantaneous sound pressure :math:`p`.
    :param reference_pressure: Reference value :math:`p_0`.
    :param axis: Axis.

    .. math:: L_{p,T} = L_{p,eqT} = 10.0 \\log_{10}{ \\left( \\frac{\\frac{1}{T} \\int_{t_1}^{t_2} p^2 (t) \\mathrm{d} t  }{p_0^2} \\right)}

    See section 2.3.
    """
    return 10.0 * np.log10( (pressure**2.0).mean(axis=axis) / reference_pressure**2.0)


def max_sound_pressure_level(pressure, reference_pressure=REFERENCE_PRESSURE, axis=-1):
    """
    Maximum time-averaged sound pressure level :math:`L_{F,max}` in dB.

    :param pressure: Instantaneous sound pressure :math:`p`.
    :param reference_pressure: Reference value :math:`p_0`.
    :param axis: Axis.

    .. math:: \mathrm{max}{(L_{p})}

    """
    return sound_pressure_level(pressure, reference_pressure=reference_pressure).max(axis=axis)

def peak_sound_pressure(pressure, axis=-1):
    """
    Peak sound pressure :math:`p_{peak}` is the greatest absolute sound pressure during a certain time interval.

    :param pressure: Instantaneous sound pressure :math:`p`.
    :param axis: Axis.

    .. math:: p_{peak} = \\mathrm{max}(|p|)

    """
    return np.abs(pressure).max(axis=axis)


def peak_sound_pressure_level(pressure, reference_pressure=REFERENCE_PRESSURE, axis=-1):
    """
    Peak sound pressure level :math:`L_{p,peak}` in dB.

    :param pressure: Instantaneous sound pressure :math:`p`.
    :param reference_pressure: Reference value :math:`p_0`.
    :param axis: Axis.

    .. math:: L_{p,peak} = 10.0 \\log \\frac{p_{peak}^2.0}{p_0^2}

    """
    return 10.0 * np.log10 (peak_sound_pressure(pressure, axis=axis)**2.0 / reference_pressure**2.0)


REFERENCE_SOUND_EXPOSURE = 4.0e-10
"""
Reference value of the sound exposure :math:`E_0` is :math:`4 \cdot 10^{-12} \\mathrm{Pa}^2\\mathrm{s}`.
"""

def sound_exposure(pressure, fs, axis=-1):
    """
    Sound exposure :math:`E_{T}`.

    :param pressure: Instantaneous sound pressure :math:`p`.
    :param fs: Sample frequency :math:`f_s`.
    :param axis: Axis.

    .. math:: E_T = \\int_{t_1}^{t_2} p^2 (t) \\mathrm{d}t

    """
    return (pressure**2.0/fs).sum(axis=axis)

def sound_exposure_level(pressure, fs, reference_sound_exposure=REFERENCE_SOUND_EXPOSURE, axis=-1):
    """
    Sound exposure level :math:`L_{E,T}` in dB.

    :param pressure: Instantaneous sound pressure :math:`p`.
    :param fs: Sample frequency :math:`f_s`.
    :param sound_exposure: Sound exposure :math:`E_{T}`.
    :param reference_sound_exposure: Reference value :math:`E_{0}`

    .. math:: L_{E,T} = 10 \\log_{10}{ \\frac{E_T}{E_0}  }

    """
    return 10.0 * np.log10(sound_exposure(pressure, fs, axis=axis)/reference_sound_exposure)


REFERENCE_POWER = 1.0e-12
"""
Reference value of the sound power :math:`P_0` is 1 pW.
"""

def sound_power_level(power, reference_power=REFERENCE_POWER):
    """
    Sound power level :math:`L_{W}`.

    :param power: Sound power :math:`P`.
    :param reference_power: Reference sound power :math:`P_0`.

    .. math:: 10 \\log_{10}{ \\frac{P}{P_0}  }

    """
    return 10.0 * np.log10(power/reference_power)


def sound_energy(power, axis=-1):
    """
    Sound energy :math:`J`..

    :param power: Sound power :math:`P`.

    .. math:: J = \\int_{t_1}^{t_2} P(t) \\mathrm{d} t
    """
    return power.sum(axis=axis)


REFERENCE_ENERGY = 1.0e-12
"""
Reference value of the sound energy :math:`J_0` is 1 pJ.
"""

def sound_energy_level(energy, reference_energy=REFERENCE_ENERGY):
    """
    Sound energy level L_{J} in dB.

    :param energy: Sound energy :math:`J`.
    :param reference_energy: Reference sound energy :math:`J_0`.

    .. math:: L_{J} = 10 \\log_{10}{ \\frac{J}{J_0} }

    """
    return np.log10( energy/reference_energy )

def sound_intensity(pressure, velocity):
    """
    Sound intensity :math:`\\mathbf{i}`.

    :param pressure: Sound pressure :math:`p(t)`.
    :param velocity: Particle velocity :math:`\\mathbf{u}(t)`.

    .. math:: \\mathbf{i} = p(t) \cdot \\mathbf{u}(t)

    """
    return pressure * velocity


REFERENCE_INTENSITY = 1.0e-12
"""
Reference value of the sound intensity :math:`I_0` is :math:`\\mathrm{1 pW/m^2}`.
"""

def time_averaged_sound_intensity(intensity, axis=-1):
    """
    Time-averaged sound intensity :math:`\\mathbf{I}_T`.

    :param intensity: Sound intensity :math:`\\mathbf{i}`.
    :param axis: Axis.

    .. math:: \\mathbf{I}_T = \\frac{1}{T} \\int_{t_1}^{t_2} \\mathbf{i}(t)

    """
    return intensity.mean(axis=axis)


def time_averaged_sound_intensity_level(time_averaged_sound_intensity, reference_intensity=REFERENCE_INTENSITY, axis=-1):
    """
    Time-averaged sound intensity level :math:`L_{I,T}`.

    :param time_averaged_sound_intensity: Time-averaged sound intensity :math:`\\mathbf{I}_T`.
    :param reference_intensity: Reference sound intensity :math:`I_0`.

    .. math:: L_{I,T} = 10 \\log_{10} { \\frac{|\\mathbf{I}_T|}{I_0} }

    """
    return 10.0 * np.log10( np.linalg.norm(time_averaged_sound_intensity,axis=axis) / reference_intensity )


def normal_time_averaged_sound_intensity(time_averaged_sound_intensity, unit_normal_vector):
    """
    Normal time-averaged sound intensity :math:`I_{n,T}`.

    :param time_averaged_sound_intensity: Time-averaged sound intensity :math:`\\mathbf{I}_T`.
    :param unit_normal_vector: Unit normal vector :math:`\\mathbf{n}`.

    .. math:: I_{n,T} = \\mathbf{I}_T \\cdot \\mathbf{n}

    """
    return time_averaged_sound_intensity.dot(unit_normal_vector)


def normal_time_averaged_sound_intensity_level(normal_time_averaged_sound_intensity, reference_intensity=REFERENCE_INTENSITY):
    """
    Normal time-averaged sound intensity level :math:`L_{In,T}` in dB.

    :param normal_time_averaged_sound_intensity: Normal time-averaged sound intensity :math:`I{n,T}`.
    :param reference_intensity: Reference sound intensity :math:`I_0`.

    .. math:: I_{n,T} = 10 \\log_{10} { \\frac{|I_{n,T}|}{I_0}}

    """
    return 10.0 * np.log10( np.abs(normal_time_averaged_sound_intensity / reference_intensity) )



"""
Bands
=====

"""

from __future__ import division

import numpy as np
#from acoustics.decibel import dbsum
import acoustics
from acoustics.standards.iec_61672_1_2013 import NOMINAL_OCTAVE_CENTER_FREQUENCIES as OCTAVE_CENTER_FREQUENCIES
from acoustics.standards.iec_61672_1_2013 import NOMINAL_THIRD_OCTAVE_CENTER_FREQUENCIES as THIRD_OCTAVE_CENTER_FREQUENCIES

def octave(first, last):
    """
    Generate a Numpy array for central frequencies of octave bands.

    There are more information on how to calculate 'real' bands in
    http://blog.prosig.com/2006/02/17/standard-octave-bands/

    Parameters
    ----------
    first : scalar
        First octave centerfrequency.

    last : scalar
        Last octave centerfrequency.

    Returns
    -------
    octave_bands : array
        An array of centerfrequency octave bands.
    """
    #octave_bands = OCTAVE_CENTER_FREQUENCIES
    #low = np.where(octave_bands == first)[0]
    #high = np.where(octave_bands == last)[0]
    #return octave_bands[low: high+1]
    return acoustics.signal.OctaveBand(fstart=first, fstop=last, fraction=1).nominal


def octave_low(first, last):
    return octave(first, last)/np.sqrt(2.0)
    #return acoustics.signal.OctaveBand(fstart=first, fstop=last, fraction=1).lower


def octave_high(first, last):
    return octave(first, last)*np.sqrt(2.0)
    #return acoustics.signal.OctaveBand(fstart=first, fstop=last, fraction=1).upper


def third(first, last):
    """
    Generate a Numpy array for central frequencies of third octave bands.

    Parameters
    ----------
    first : scalar
       First third octave centerfrequency.

    last : scalar
        Last third octave centerfrequency.

    Returns
    -------
    octave_bands : array
        An array of centerfrequency third octave bands.
    """
    #third_oct_bands = THIRD_OCTAVE_CENTER_FREQUENCIES
    #low = np.where(third_oct_bands == first)[0]
    #high = np.where(third_oct_bands == last)[0]
    #return third_oct_bands[low: high+1]
    return acoustics.signal.OctaveBand(fstart=first, fstop=last, fraction=3).nominal


def third_low(first, last):
    return third(first, last)/2.0**(1.0/6.0)
    #return acoustics.signal.OctaveBand(fstart=first, fstop=last, fraction=3).lower


def third_high(first, last):
    return third(first, last)*2.0**(1.0/6.0)
    #return Octaveband(fstart=first, fstop=last, fraction=3).upper

    
def third2oct(levels, axis=None):
    """
    Calculate Octave levels from third octave levels.
    
    :param levels: Array containing third octave levels.
    :type: :class:`np.ndarray`
    :param axis: Axis over which to perform the summation. 
    :type axis: :class:`int`
    
    :returns: Third octave levels
    :rtype: :class:`np.ndarray`
    
    .. note:: The number of elements along the summation axis should be a factor of 3.
    """
    
    levels = np.array(levels)
    axis = axis if axis is not None else levels.ndim - 1
    
    try:
        assert(levels.shape[axis]%3 == 0)
    except AssertionError:    
        raise ValueError("Wrong shape.")
    shape = list(levels.shape)
    shape[axis] = shape[axis] // 3
    shape.insert(axis+1, 3)
    levels = np.reshape(levels, shape)
    return np.squeeze(acoustics.decibel.dbsum(levels, axis=axis+1))

def _check_band_type(freqs):
    """Check if an array contains octave or third octave bands values sorted
    or unsorted.
    """
    octave_bands = octave(16, 16000)
    third_oct_bands = third(12.5, 20000)

    def _check_sort(freqs, bands):
        index = np.where(np.in1d(bands, freqs))[0]
        band_pos = index - index[0]
        if (band_pos == np.arange(band_pos.size)).all():
            sorted = True
        else:
            sorted = False
        return sorted

    if np.in1d(freqs, octave_bands).all() == True:
        is_sorted = _check_sort(freqs, octave_bands)
        if is_sorted is True:
            band_type = 'octave'
        else:
            band_type = 'octave-unsorted'
    elif np.in1d(freqs, third_oct_bands).all() == True:
        is_sorted = _check_sort(freqs, third_oct_bands)
        if is_sorted is True:
            band_type = 'third'
        else:
            band_type = 'third-unsorted'
    else:
        band_type = None

    return band_type



"""
ISO 1996-2:2007

ISO 1996-2:2007 describes how sound pressure levels can be determined by direct measurement, 
by extrapolation of measurement results by means of calculation, or exclusively by calculation, 
intended as a basis for assessing environmental noise. 

"""
import itertools
import numpy as np
import pandas as pd
from scipy.signal import welch, hanning
from scipy.stats import linregress
import matplotlib.pyplot as plt
from acoustics.decibel import dbsum
from acoustics.standards.iso_tr_25417_2007 import REFERENCE_PRESSURE
import weakref
from tabulate import tabulate


TONE_WITHIN_PAUSE_CRITERION_DB = 6.0
"""A tone may exist when the level of any line in the noise pause is 6 dB or more about...."""

TONE_BANDWIDTH_CRITERION_DB = 3.0
"""Bandwidth of the detected peak."""

TONE_LINES_CRITERION_DB = 6.0
"""All lines with levels within 6 dB of the maximum level are classified as tones."""

TONE_SEEK_CRITERION = 1.0
"""Tone seek criterion."""

REGRESSION_RANGE_FACTOR = 0.75
"""Range of regression is usually +/- 0.75 critical bandwidth."""

_WINDOW_CORRECTION = {'hanning' : -1.8,
                      }

def window_correction(window):
    """Correction to be applied to :math:`L_{pt}` due to use of window."""
    try:
        return _WINDOW_CORRECTION[window]
    except KeyError:
        raise ValueError("Window correction is not available for specified window.")


def critical_band(frequency):
    """Bandwidth of critical band of frequency.
    
    :param frequency: Center frequency of tone.
    :returns: (bandwidth, center, lower, upper) of band.
    
    """
    if isinstance(frequency, np.ndarray):
        center[frequency<50.0] = 50.0
    else:
        center = 50.0 if frequency < 50 else frequency

    bandwidth = (center>500.0)*(center*0.20) + (center<=500.0)*100.0
    
    upper = center + bandwidth/2.0
    lower = center - bandwidth/2.0
    
    return center, lower, upper, bandwidth


def tones_level(tone_levels):
    """Total sound pressure level of the tones in a critical band given the level of each of the tones.
    
    .. math L_{pt} = 10 \log_{10}{\sum 10^{L_{pti}/10}}
    
    See equation C.1 in section C.2.3.1.
    """
    return dbsum(tone_levels)


def masking_noise_level(noise_lines, frequency_resolution, effective_analysis_bandwidth):
    """Masking noise level :math:`L_{pn}`
    
    :param noise_lines: Masking noise lines. See :func:`masking_noise_lines`.
    :param frequency_resolution: Frequency resolution :math:`\Delta f`.
    :param effective_analysis_bandwidth: Effective analysis bandwidth :math:`B`.
    
    .. math:: L_{pn} = 10 \log_{10}{\sum 10^{L_n/10}} + 10 \log_{10}{\frac{\Delta f}{B}}
    
    See equation C.11 in section C.4.4.
    
    """
    return dbsum(noise_lines) + 10.0*np.log10(frequency_resolution/effective_analysis_bandwidth)


def masking_noise_lines(levels, line_classifier, center, bandwidth, regression_range_factor):
    """Determine masking noise level lines using regression line. Returns array of :math:`L_n`.
    
    :param levels: Levels as function of frequency.
    :type levels: :class:`pd.Series`.
    :param lines_classifier: Categorical indicating what each line is.
    :param center: Center frequency. 
    :param bandwidth: bandwidth of critical band.
    :param regression_range_factor: Range factor.
    :returns: (Array with masking noise lines, slope, intercept).
    """
    slicer = slice(center-bandwidth*regression_range_factor, center+bandwidth*regression_range_factor)
    levels = levels[slicer]
    frequencies = levels.index
    regression_levels = levels[line_classifier=='noise']
    slope, intercept = linregress(x=regression_levels.index, y=regression_levels)[0:2]
    levels_from_regression = slope * frequencies + intercept
    return levels_from_regression, slope, intercept


def tonal_audibility(tones_level, masking_noise_level, center):
    """Tonal audibility.
    
    :param tones_level: Total sound pressure level of the tones in the critical band :math:`L_{pt}.
    :param masking_noise_level: Total sound pressure level of the masking noise in the critical band :math:`L_{pn}.
    :param center: Center frequency of the critical band :math:`f_c`.
    :returns: Tonal audibility :math:`\Delta L_{ta}`
    
    .. math:: \Delta L_{ta} = L_{pt} - L_{pn} + 2 + \log_{10}{1 + \left(\frac{f_c}{502}\right)^{2.5}}
    
    See equation C.3. in section C.2.4.
    """
    return tones_level - masking_noise_level + 2.0 + np.log10(1.0+(center/502.0)**(2.5))


def tonal_adjustment(tonal_audibility):
    """Adjustment :math:`K`. 
    
    :param tonal_audibility: Tonal audibility :math:`L_{ta}`.
    
    See equations C.4, C.5 and C.6 in section C.2.4.
    """
    if tonal_audibility > 10.0:
        return 6.0
    elif tonal_audibility < 4.0:
        return 0.0
    else:
        return tonal_audibility - 4.0
    

class Tonality(object):
    """Perform assessment of audibility of tones in noise.
    
    Objective method for assessing the audibility of tones in noise.
    """
    
    def __init__(self, signal, sample_frequency, window='hanning', 
                 reference_pressure=REFERENCE_PRESSURE,
                 tsc=TONE_SEEK_CRITERION,
                 regression_range_factor=REGRESSION_RANGE_FACTOR,
                 nbins=None,
                 force_tone_without_pause=False,
                 force_bandwidth_criterion=False,
                ):
        
        self.signal = signal
        """Samples in time-domain."""
        self.sample_frequency = sample_frequency
        """Sample frequency."""
        self.window = window
        """Window to be used."""
        self.reference_pressure = reference_pressure
        """Reference sound pressure."""
        self.tsc = tsc
        """Tone seeking criterium."""
        self.regression_range_factor = regression_range_factor
        """Regression range factor."""
        self.nbins = nbins
        """Amount of frequency nbins to use. See attribute `nperseg` of :func:`scipy.signal.welch`."""
        
        self._noise_pauses = list()
        """Private list of noise pauses that were determined or assigned."""
        self._spectrum = None
        """Power spectrum as function of frequency."""
        
        self.force_tone_without_pause = force_tone_without_pause
        self.force_bandwidth_criterion = force_bandwidth_criterion


    @property
    def noise_pauses(self):
        """Noise pauses that were determined."""
        for noise_pause in self._noise_pauses:
            yield noise_pause
    
    @property
    def tones(self):
        """Tones that were determined."""
        for noise_pause in self.noise_pauses:
            if noise_pause.tone is not None:
                yield noise_pause.tone
        
    @property
    def critical_bands(self):
        """Critical bands that were determined. A critical band is determined for each tone."""
        for tone in self.tones:
            yield tone.critical_band

    @property
    def spectrum(self):
        """Power spectrum of the input signal.
        """
        if self._spectrum is None:
            nbins = self.nbins
            if nbins is None:
                nbins = self.sample_frequency
            nbins //= 1 # Fix because of bug in welch with uneven nbins
            f, p = welch(self.signal, fs=self.sample_frequency, nperseg=nbins, window=self.window, detrend=False, scaling='spectrum')
            self._spectrum = pd.Series(10.0*np.log10(p / self.reference_pressure**2.0), index=f)
        return self._spectrum 
    
        
    @property
    def frequency_resolution(self):
        """Frequency resolution.
        """
        df = np.diff(np.array(self.spectrum.index)).mean()
        return df
        #return 1.0 / self.sample_frequency
    
    
    @property
    def effective_analysis_bandwidth(self):
        """Effective analysis bandwidth.
        
        In the case of the Hanning window
        
        .. math:: B_{eff} = 1.5 \Delta f
        
        with \Delta f the :attr:`frequency_resolution`.
        
        C.2.2: Note 1.
        """
        if self.window == 'hanning':
            return 1.5 *  self.frequency_resolution
        else:
            raise ValueError()
    
    def _set_noise_pauses(self, noise_pauses):
        """Manually set noise pauses. Expects iterable of tuples."""
        self._noise_pauses = [NoisePause(start, end) for start, end in noise_pauses]
        return self

    def determine_noise_pauses(self, end=None):
        """Determine noise pauses. The determined noise pauses are available in :attr:`noise_pause_ranges`.
        
        Noise pauses are search for using :func:`noise_pause_seeker`.
        """
        self._set_noise_pauses(noise_pause_seeker(np.array(self.spectrum[:end]), self.tsc))
        return self
        
    def _construct_line_classifier(self):
        """Set values of line classifier."""
        
        # Build classifier.
        levels = self.spectrum
        
        categories = ['noise', 'start', 'end', 'neither', 'tone']
        self.line_classifier = pd.Series(pd.Categorical(['noise']*len(levels), categories=categories), index=levels.index)
        
        # Add noise pauses
        for noise_pause in self.noise_pauses:
            # Mark noise pause start and end.
            self.line_classifier.iloc[noise_pause.start] = 'start'
            self.line_classifier.iloc[noise_pause.end] = 'end'
            # Mark all other lines within noise pause as neither tone nor noise.
            self.line_classifier.iloc[noise_pause.start+1:noise_pause.end] = 'neither' # Half-open interval
        
        # Add tone lines
        for tone in self.tones:
            self.line_classifier.iloc[tone._tone_lines] = 'tone'
            
        return self

    def _determine_tones(self):
        """Analyse the noise pauses for tones. The determined tones are available via :attr:`tones`.
        Per frequency line results are available via :attr:`line_classifier`.
        """
        levels = self.spectrum
        
        # First we need to check for the tones.
        for noise_pause in self.noise_pauses:            
            # Determine the indices of the tones in a noise pause
            tone_indices, bandwidth_for_tone_criterion = determine_tone_lines(levels, 
                                                                              self.frequency_resolution, 
                                                                              noise_pause.start, 
                                                                              noise_pause.end,
                                                                              self.force_tone_without_pause,
                                                                              self.force_bandwidth_criterion,
                                                                              )
            # If we have indices, ...
            if np.any(tone_indices):
                # ...then we create a tone object.
                noise_pause.tone = create_tone(levels, tone_indices, bandwidth_for_tone_criterion, weakref.proxy(noise_pause))
        return self

    def _determine_critical_bands(self):
        """Put a critical band around each of the determined tones."""
        
        for tone in self.tones:
            critical_band = self.critical_band_at(tone.center)
            tone.critical_band = critical_band
            critical_band.tone = weakref.proxy(tone)
        return self
    
    def analyse(self):
        """Analyse the noise pauses for tones and put critical bands around each of these tones. 
        The tones are available via :attr:`tones` and the critical bands via :attr:`critical_bands`.
        Per frequency line results are available via :attr:`line_classifier`.
        """
        

        # Determine tones. Puts noise pause starts/ends in classier as well as tone lines
        # and lines that are neither tone nor noise.
        self._determine_tones()
        # Construct line classifier
        self._construct_line_classifier()
        # Determine critical bands.
        self._determine_critical_bands()
        return self
    
    def critical_band_at(self, frequency):
        """Put at a critical band at `frequency`.
        In order to use this function :attr:`line_classifier` needs to be available, 
        which means :meth:`analyse` needs to be used first.
        """
        return create_critical_band(self.spectrum, self.line_classifier, frequency, 
                                    self.frequency_resolution, 
                                    self.effective_analysis_bandwidth, 
                                    self.regression_range_factor, self.window)


    def plot_spectrum(self):
        """Plot power spectrum."""
        spectrum = self.spectrum
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.plot(spectrum.index, spectrum)
        ax.set_xlabel('f in Hz')
        ax.set_ylabel('L in dB')
        return fig
    
    @property
    def dominant_tone(self):
        """Most dominant_tone tone.
        
        The most dominant_tone tone is the tone with the highest tonal audibility :math:`L_{ta}`.
        """
        try:
            return sorted(self.tones, key=lambda x: x.critical_band.tonal_audibility, reverse=True)[0]
        except IndexError:
            return None
        #elif self.tones:
            #raise RuntimeError("Tones were determined but critical bands were not...?")
        #elif self.noise_pauses:
            #raise ValueError("Need to run analysis first.")
        #else:
            #raise ValueError("Need to define/search noise pauses first.")
    
    
    def plot_results(self, noise_pauses=False, tones=True, critical_bands=True):
        """Plot overview of results."""

        df = self.frequency_resolution
        #fig = self.plot_spectrum()
        #ax = fig.axes[0]

        levels = self.spectrum
        
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.plot(levels.index, levels)
        ax.set_xlabel("$f$ in Hz")
        ax.set_ylabel("$L$ in dB")
        
        if noise_pauses:
            for pause in self.noise_pauses:
                ax.axvspan(pause.start*df, pause.end*df, color='green', alpha=0.05)
        
        if tones:
            for tone in self.tones:
                ax.axvline(tone.center, color='black', alpha=0.05)
        
        if critical_bands:
            for band in self.critical_bands:
                ax.axvspan(band.start, band.end, color='yellow', alpha=0.05)

        band = self.dominant_tone.critical_band
            #ax.avxspan(band.start, band.end, color='blue')
        ax.axvline(band.start, color='red', linewidth=0.1)
        ax.axvline(band.end, color='red', linewidth=0.1)

        # Limit xrange
        if noise_pauses: 
            _items = list(self.noise_pauses)
        elif critical_bands: 
            _items = list(self.critical_bands)
        #span = IntervalTree(Interval(pause.start, pause.end) for pause in _items).range()
        #ax.set_xlim(span.begin, span.end)
        ax.set_xlim(min(item.start for item in _items), max(item.end for item in _items))
        return fig
    
    
    def overview(self):
        """Print overview of results."""
        try:
            cb = self.dominant_tone.critical_band
        except AttributeError:
            raise ValueError("Cannot show overview (yet). No tones have been determined.")
        
        tones = [("Tone", "{:4.1f} Hz: {:4.1f} dB".format(tone.center, tone.tone_level)) for tone in self.tones]
        
        table = [("Critical band", "{:4.1f} to {:4.1f} Hz".format(cb.start, cb.end)),
                 ("Masking noise level $L_{pn}$", "{:4.1f} dB".format(cb.masking_noise_level)),
                 ("Tonal level $L_{pt}$", "{:4.1f} dB".format(cb.total_tone_level)),
                 ("Dominant tone", "{:4.1f} Hz".format(cb.tone.center)),
                 ("3 dB bandwidth of tone", "{:2.1f}% of {:4.1f}".format(cb.tone.bandwidth_3db/cb.bandwidth*100.0, cb.bandwidth)),
                 ("Tonal audibility $L_{ta}$", "{:4.1f} dB".format(cb.tonal_audibility)),
                 ("Adjustment $K_{t}$", "{:4.1f} dB".format(cb.adjustment)),
                 ("Frequency resolution", "{:4.1f} Hz".format(self.frequency_resolution)),
                 ("Effective analysis bandwidth", "{:4.1f} Hz".format(self.effective_analysis_bandwidth)),
            ]
        table += tones
        return tabulate(table)
              
    def results_as_dataframe(self):
        """Return results in dataframe."""
        data = ((tone.center, tone.tone_level, tone.bandwidth_3db, tone.critical_band.start, 
                 tone.critical_band.end, tone.critical_band.bandwidth, 
                 tone.critical_band.regression_slope, tone.critical_band.regression_intercept, 
                 tone.critical_band.masking_noise_level, tone.critical_band.total_tone_level, 
                 tone.critical_band.tonal_audibility, tone.critical_band.adjustment) for tone in self.tones)
        columns = ['center', 'tone_level', 'bandwidth_3db', 'critical_band_start', 'critical_band_end', 
                   'critical_band_bandwidth', 'regression_slope', 'regression_intercept', 'masking_noise_level', 
                   'total_tone_level', 'tonal_audibility', 'adjustment']
        return pd.DataFrame(list(data), columns=columns)
        
        
 
class NoisePause(object):
    
    def __init__(self, start, end, tone=None):
        self.start = start
        self.end = end
        self.tone = tone
        
    def __str__(self):
        return "(start={},end={})".format(self.start, self.end)
    
    def __repr__(self):
        return "NoisePause{}".format(str(self))

    def __iter__(self):
        yield self.start
        yield self.stop

    def _repr_html_(self):
        table = [("Start", self.start),
                 ("End", self.end)]
        return tabulate(table, tablefmt="html")


def create_tone(levels, tone_lines, bandwidth_for_tone_criterion, noise_pause):
    """Create an instance of Tone."""
    
    center = levels.iloc[tone_lines].argmax()
    tone_level = tones_level(levels.iloc[tone_lines])
    return Tone(center, tone_lines, tone_level, noise_pause, bandwidth_for_tone_criterion)


class Tone(object):
    """Tone."""
    
    def __init__(self, center, tone_lines, tone_level, noise_pause, bandwidth_3db, critical_band=None):
        self.center = center
        self._tone_lines = tone_lines
        self.tone_level = tone_level
        self.noise_pause = noise_pause
        self.bandwidth_3db = bandwidth_3db
        self.critical_band = critical_band

    def __str__(self):
        return "(center={:4.1f}, levels={:4.1f})".format(self.center, self.tone_level)

    def __repr__(self):
        return "Tone{}".format(str(self))

    def _repr_html_(self):
        table = [("Center frequency", "{:4.1f} Hz".format(self.center)),
                 ("Tone level", "{:4.1f} dB".format(self.tone_level))]
        return tabulate(table, tablefmt='html')


def create_critical_band(levels, line_classifier, frequency, frequency_resolution,
                         effective_analysis_bandwidth, regression_range_factor, window, tone=None):
    """Create an instance of CriticalBand."""
    
    center, start, end, bandwidth = critical_band(frequency)
    
    # Masking noise lines
    noise_lines, regression_slope, regression_intercept = masking_noise_lines(levels, line_classifier, center, bandwidth, regression_range_factor)
    # Masking noise level   
    noise_level = masking_noise_level(noise_lines, frequency_resolution, effective_analysis_bandwidth)
    # Total tone level
    tone_lines = levels[line_classifier=='tone'][start:end]
    tone_level = tones_level(tone_lines) - window_correction(window)
    # Tonal audibility
    audibility = tonal_audibility(tone_level, noise_level, center)
    # Adjustment Kt
    adjustment = tonal_adjustment(audibility)
    
    return CriticalBand(center, start, end, bandwidth, 
                        regression_range_factor, regression_slope, regression_intercept,
                        noise_level, tone_level, audibility, adjustment, tone)


class CriticalBand(object):
    
    def __init__(self, center, start, end, bandwidth, 
                 regression_range_factor, regression_slope, regression_intercept, 
                 noise_level, tone_level, audibility,
                 adjustment, tone=None):
                 
        self.center = center
        """Center frequency of the critical band."""
        self.start = start
        """Lower band-edge frequency of the critical band."""
        self.end = end
        """Upper band-edge frequency of the critical band."""
        self.bandwidth = bandwidth
        """Bandwidth of the critical band."""
        self.regression_range_factor = regression_range_factor
        """Range of regression factor. See also :attr:`REGRESSION_RANGE_FACTOR`."""        
        self.regression_slope = regression_slope
        """Linear regression slope."""
        self.regression_intercept = regression_intercept
        """Linear regression intercept."""
        self.masking_noise_level = noise_level
        """Masking noise level :math:`L_{pn}`."""
        self.total_tone_level = tone_level
        """Total tone level :math:`L_{pt}`."""
        self.tonal_audibility = audibility
        """Tonal audibility :math:`L_{ta}`."""
        self.adjustment = adjustment
        """Adjustment :math:`K_{t}`."""
        self.tone = tone

    def __str__(self):
        return "(center={:4.1f}, bandwidth={:4.1f}, tonal_audibility={:4.1f}, adjustment={:4.1f}".format(self.center,
                                                                                                         self.bandwidth,
                                                                                                         self.tonal_audibility,
                                                                                                         self.adjustment)
 
    def __repr__(self):
        return "CriticalBand{}".format(str(self))

    def _repr_html_(self):
        
        table = [
            ("Center frequency", "{:4.1f} Hz".format(self.center)),
            ("Start frequency", "{:4.1f} Hz".format(self.start)),
            ("End frequency", "{:4.1f} Hz".format(self.end)),
            ("Bandwidth", "{:4.1f} Hz".format(self.bandwidth)),
            ("Regression factor", "{:4.1f}".format(self.regression_range_factor)),
            ("Regression slope", "{:4.1f}".format(self.regression_slope)),
            ("Regression intercept", "{:4.1f}".format(self.regression_intercept)),
            ("Masking noise level", "{:4.1f} dB".format(self.masking_noise_level)),
            ("Total tone level", "{:4.1f} dB".format(self.total_tone_level)),
            ("Tonal audibility $L_{ta}$", "{:4.1f} dB".format(self.tonal_audibility)),
            ("Adjustment $K_{t}$", "{:4.1f} dB".format(self.adjustment)),
            ]
        
        return tabulate(table, tablefmt='html')

    
#----------Noise pauses----------------------------  

def _search_noise_pauses(levels, tsc):
    pauses = list()
    possible_start = None
    for i in range(2, len(levels)-2):
            if (levels[i] - levels[i-1]) >= tsc and (levels[i-1] - levels[i-2]) < tsc:
                possible_start = i
            if (levels[i] - levels[i+1]) >= tsc and (levels[i+1] - levels[i+2]) < tsc:
                if possible_start:
                    pauses.append((possible_start, i))
                    possible_start = None
    return pauses
    
    
def noise_pause_seeker(levels, tsc):
    """Given the levels of a spectrum and a tone seeking criterium this top level function seeks possible noise pauses.
    
    :param levels: Spectral levels.
    :param df: Frequency resolution.
    :param tsc: Tone seeking criterium.
    
    Possible start and end indices of noise pauses are determined using :func:`possible_noise_pauses.
    Then, only those that correspond to the smallest intervals that do not overlap other intervals are kept.
    """    
    n = len(levels)
    forward_pauses = _search_noise_pauses(levels, tsc)
    backward_pauses = _search_noise_pauses(levels[::-1], tsc)
    backward_pauses = [(n-1-start, n-1-end) for end, start in reversed(backward_pauses)]
    possible_pauses = sorted(list( set(forward_pauses) & set(backward_pauses) ) )
    return possible_pauses

#------------------- Tone seeking--------------------

def determine_tone_lines(levels, df, start, end, force_tone_without_pause=False, force_bandwidth_criterion=False):
    """Determine tone lines in noise pause.
      
    :param levels: Series with levels as function of frequency.
    :param df: Frequency resolution.
    :param start: Index of noise pause start.
    :param end: Index of noise pause end.
    
    :returns: Array with indices of tone lines in noise pause.
    
    """
    # Noise pause range object
    npr = slice(start, end+1)

    # Return values
    tone_indices = np.array([])
    bandwidth_for_tone_criterion = None
    
    # Levels but with integeres as indices instead of frequencies.
    # Benefit over np.array is that the index is maintained when the object is sliced.
    levels_int = levels.reset_index(drop=True)
    
    # If any of the lines is six 6 dB above. See section C.4.3.
    if np.any((levels.iloc[npr] >= TONE_WITHIN_PAUSE_CRITERION_DB + levels.iloc[start-1]) & 
              (levels.iloc[npr] >= TONE_WITHIN_PAUSE_CRITERION_DB + levels.iloc[end+1])) or force_tone_without_pause:

        # Indices of values that are within -3 dB point.
        indices_3db = np.nonzero(levels.iloc[npr] >= levels.iloc[npr].max() - TONE_BANDWIDTH_CRITERION_DB)[0]
        # -3 dB bandwidth
        bandwidth_for_tone_criterion = (indices_3db.max()-indices_3db.min()) * df
        # Frequency of tone.
        tone_center_frequency = levels.iloc[npr].argmax()
        tone_center_index = levels.reset_index(drop=True).iloc[npr].argmax()
        # Critical band
        _, _, _, critical_band_bandwidth = critical_band(tone_center_frequency)
        
        # Fullfill bandwidth criterion? See section C.4.3
        if (bandwidth_for_tone_criterion < 0.10 * critical_band_bandwidth) or force_bandwidth_criterion:
            # All values within 6 decibel are designated as tones.
            tone_indices = (levels_int.iloc[npr][ levels_int.iloc[npr] >= levels_int.iloc[npr].max() - TONE_LINES_CRITERION_DB ]).index.get_values()
            
    return tone_indices, bandwidth_for_tone_criterion



"""
Power
=====

"""

from __future__ import division
import numpy as np


def lw_iso3746(LpAi, LpAiB, S, alpha, surfaces):
    """
    Calculate sound power level according to ISO 3746:2010.
    
    :param LpAi: Sound pressure levels of the source :math:`L_{pAi}`.
    :type LpAi: :class:`np.ndarray`
    :param LpAiB: Background noise sound pressure levels :math:`L_{pAiB}`.
    :type LpAiB: :class:`np.ndarray`
    :param S: Area in square meters of the measurement surface :math:`S`.
    :type S: :class:`float`
    :param alpha: Absorption coefficients of the room :math:`\\alpha`. 
    :type alpha: :class:`np.ndarray`
    :param surfaces: Room surfaces.
    :type surfaces: :class:`np.ndarray`

    :returns: Sound power level :math:`L_{w}`.
    """
    LpA = 10.0 * np.log10(np.sum(10.0**(0.1*LpAi))/LpAi.size)
    LpAB = 10.0 * np.log10(np.sum(10.0**(0.1*LpAiB))/LpAiB.size)
    deltaLpA = LpA - LpAB

    if deltaLpA > 10.0:
        k_1a = 0.0
    elif 3.0 <= deltaLpA <= 10.0:
        k_1a = -10.0 * np.log10(1.0 - 10.0**(-0.1*deltaLpA))
    else:
        # This should alert to user because poor condition of the measurement.
        k_1a = 3.0

    S0 = 1.0
    Sv = np.sum(surfaces)
    alpha_mean = np.average(alpha, axis=0, weights=surfaces)
    A = alpha_mean * Sv

    k_2a = 10.0 * np.log10(1.0 + 4.0 * S / A)

    LpA_mean = LpA - k_1a - k_2a
    L_WA = LpA_mean + 10.0 * np.log10(S / S0)
    return L_WA
